from collections.abc import Iterable

from boa3_test.test_drive.model.interface.itransactionobject import ITransactionObject
from boa3_test.test_drive.model.invoker import invokeresult
from boa3_test.test_drive.model.invoker.neoinvoke import NeoInvoke


class NeoBatchInvoke(ITransactionObject):
    def __init__(self, invoke: NeoInvoke, tx_pos: int, expected_result_type: type = None):
        super().__init__()
        self._invoke: NeoInvoke = invoke
        self._tx_pos: int = tx_pos
        self._result = invokeresult.NOT_EXECUTED

        if not isinstance(expected_result_type, type):
            expected_result_type = None
        self._expected_result_type = expected_result_type

    @property
    def invoke(self) -> NeoInvoke:
        return self._invoke

    @property
    def result(self):
        if hasattr(self._result, 'copy'):
            result = self._result.copy()
        else:
            result = self._result

        if self._expected_result_type is not None:
            if self._expected_result_type is not str and isinstance(result, str):
                from boa3.internal.neo.vm.type.String import String
                result = String(result).to_bytes()

            if self._expected_result_type in (bool, int):
                if isinstance(result, bytes):
                    from boa3.internal.neo.vm.type.Integer import Integer
                    result = Integer.from_bytes(result, signed=True)

                if self._expected_result_type is bool and isinstance(result, int) and result in (False, True):
                    result = bool(result)

            elif self._expected_result_type is bytearray and isinstance(result, bytes):
                result = bytearray(result)
            elif self._expected_result_type is tuple and isinstance(result, Iterable):
                result = tuple(result)

        return result

    def _set_data_from_match_result(self, match_groups: dict):
        super()._set_data_from_match_result(match_groups)
        if self._tx_id is not None and self._result is not invokeresult.TRANSACTION_EXECUTED:
            self._result = invokeresult.TRANSACTION_EXECUTED

    def __repr__(self):
        return f'{self._invoke.__repr__()} -> {self.tx_id.__repr__()}'



from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import neoexpresscommand as neoxp
from boa3_test.test_drive.neoxp.command.neoexpresscommand import NeoExpressCommand


def enable(account: Account) -> NeoExpressCommand:
    return neoxp.oracle.OracleEnableCommand(account)


def response(url: str, response_path: str, request_id: int = None) -> NeoExpressCommand:
    return neoxp.oracle.OracleResponseCommand(url, response_path,
                                              request_id=request_id)


def requests() -> NeoExpressCommand:
    return neoxp.oracle.OracleRequestsCommand()


def list() -> NeoExpressCommand:
    return neoxp.oracle.OracleListCommand()



from typing import Any, Self

from boa3.internal.neo import from_hex_str, utils
from boa3.internal.neo3.core.types import UInt256
from boa3.internal.neo3.vm import VMState
from boa3_test.test_drive.model.network.payloads.signer import Signer
from boa3_test.test_drive.model.network.payloads.transactionattribute import TransactionAttribute
from boa3_test.test_drive.model.network.payloads.witness import Witness
from boa3_test.test_drive.model.smart_contract.contractcollection import ContractCollection
from boa3_test.test_drive.model.smart_contract.triggertype import TriggerType
from boa3_test.test_drive.testrunner.blockchain.notification import TestRunnerNotification as Notification


class TestTransaction:
    def __init__(self, tx_hash: UInt256 | bytes, script: bytes, signers: list[Signer] = None, witnesses: list[Witness] = None):
        if isinstance(tx_hash, bytes):
            tx_hash = UInt256(tx_hash)

        self._hash: UInt256 = tx_hash
        self._script: bytes = script

        self._signers: list[Signer] = signers if signers is not None else []
        self._witnesses: list[Witness] = witnesses if witnesses is not None else []
        self._attributes: list[TransactionAttribute] = []

    def __new__(cls, *args, **kwargs):
        return super().__new__(cls)

    @property
    def hash(self) -> UInt256:
        return self._hash

    def to_json(self) -> dict[str, Any]:
        import base64
        from boa3.internal.neo.vm.type.String import String
        return {
            'hash': str(self._hash),
            'signers': [signer.to_json() for signer in self._signers],
            'witnesses': [witness.to_json() for witness in self._witnesses],
            'attributes': [attr.to_json() for attr in self._attributes if hasattr(attr, 'to_json')],
            'script': String.from_bytes(base64.b64encode(self._script))
        }

    @classmethod
    def from_json(cls, json: dict[str, Any], *args, **kwargs) -> Self:
        import base64

        if 'hash' in json and isinstance(json['hash'], str):
            tx_hash = UInt256(from_hex_str(json['hash']))
        else:
            tx_hash = UInt256.zero()

        script = base64.b64decode(json['script'])
        tx = object.__new__(cls)  # init was causing errors with inherited classes
        tx._hash = tx_hash
        tx._script = script

        tx._attributes = []
        if 'attributes' in json:
            tx._attributes = [TransactionAttribute.from_json(attr) for attr in json['attributes']]
        else:
            tx._attributes = []

        if 'signers' in json:
            signers_json = json['signers']
            if not isinstance(signers_json, list):
                signers_json = [signers_json]
            tx._signers = [Signer.from_json(js) for js in signers_json]
        else:
            tx._signers = []

        if 'witnesses' in json:
            witnesses_json = json['witnesses']
            if not isinstance(witnesses_json, list):
                witnesses_json = [witnesses_json]
            tx._witnesses = [Witness.from_json(js) for js in witnesses_json]
        else:
            tx._witnesses = []

        return tx

    def __repr__(self) -> str:
        return str(self._hash)


class TransactionExecution:
    def __init__(self):
        super().__init__()
        self._trigger = TriggerType.All
        self._vm_state = VMState.NONE
        self._gas_consumed: int = 0
        self._exception: str = None
        self._stack = []
        self._notifications: list[Notification] = []

    @property
    def trigger(self) -> TriggerType:
        return self._trigger

    @property
    def vm_state(self) -> VMState:
        return self._vm_state

    @property
    def exception(self) -> str | None:
        return self._exception

    @property
    def gas_consumed(self) -> int:
        return self._gas_consumed

    @property
    def result_stack(self) -> list:
        return self._stack.copy()

    @property
    def notifications(self) -> list[Notification]:
        return self._notifications.copy()

    @classmethod
    def from_json(cls, json: dict[str, Any], contract_collection: ContractCollection = None) -> Self:
        tx_exec = cls()

        tx_exec._trigger = TriggerType[json['trigger']]
        tx_exec._vm_state = VMState[json['vmstate']]
        tx_exec._gas_consumed = int(json['gasconsumed'])
        tx_exec._exception = str(json['exception']) if json['exception'] is not None else None
        tx_exec._stack = [utils.stack_item_from_json(value) for value in json['stack']]
        tx_exec._notifications = [Notification.from_json(notification, contract_collection) for notification in json['notifications']]

        return tx_exec



from typing import Any, Self

from boa3_test.test_drive.model.interface.itransactionobject import ITransactionObject
from boa3_test.test_drive.model.network.payloads.testtransaction import TransactionExecution
from boa3_test.test_drive.model.smart_contract.contractcollection import ContractCollection


class TestRunnerTransactionLog(ITransactionObject):
    def __init__(self):
        super().__init__()
        self._executions: list[TransactionExecution] = []

    @property
    def executions(self) -> list[TransactionExecution]:
        return self._executions.copy()

    @classmethod
    def from_json(cls, json: dict[str, Any], contract_collection: ContractCollection = None) -> Self:
        tx_log = cls()

        tx_log._executions = [TransactionExecution.from_json(execution, contract_collection) for execution in json['executions']]

        return tx_log



import ast
from abc import ABC

from boa3.internal.model.builtin.builtinsymbol import IBuiltinSymbol
from boa3.internal.model.callable import Callable
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class IBuiltinCallable(Callable, IBuiltinSymbol, ABC):
    def __init__(self,
                 identifier: str,
                 args: dict[str, Variable] = None,
                 vararg: tuple[str, Variable] | None = None,
                 kwargs: dict[str, Variable] | None = None,
                 defaults: list[ast.AST] = None,
                 return_type: IType = None,
                 deprecated: bool = False,
                 new_location: str = None
                 ):
        super().__init__(args, vararg, kwargs, defaults, return_type, deprecated=deprecated)
        self._identifier = identifier
        self._generated_opcode = None
        self._new_location = new_location
        self.defined_by_entry = False  # every builtin symbol must have this variable set as False

    @property
    def new_location(self) -> str | None:
        return self._new_location

    def set_new_location(self, new_location: str, identifier: str = None):
        """
        Set the location that is going to be shown in the deprecation warning in case if this symbol is deprecated.
        """
        from boa3.internal import constants
        if identifier is None or len(identifier.strip()) == 0:
            identifier = self._identifier

        self._new_location = constants.ATTRIBUTE_NAME_SEPARATOR.join((new_location, identifier))

    @property
    def opcode(self) -> list[tuple[Opcode, bytes]]:
        """
        Gets the opcode for the method.

        :return: the opcode and its data if exists. None otherwise.
        """
        # don't need to recalculate for every time this property is called
        if self._generated_opcode is None:
            self._generated_opcode = self._opcode
        return self._generated_opcode

    def generate_opcodes(self, code_generator):
        """
        Generate the Neo VM opcodes for the method.

        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        self.generate_internal_opcodes(code_generator)

    def generate_internal_opcodes(self, code_generator):
        """
        Generate the Neo VM opcodes for the method.

        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        pass

    def reset(self):
        # reset the opcodes to ensure the correct output when calling consecutive compilations
        self._generated_opcode = None
        self.reset_calls()

    @property
    def _opcode(self) -> list[tuple[Opcode, bytes]]:
        return []

    @property
    def identifier(self) -> str:
        return self._identifier



from boa3.internal.model.builtin.method import LenMethod
from boa3.internal.model.builtin.native.nativecontract import NativeContract
from boa3.internal.model.type.itype import IType


class LenStrMethod(LenMethod):

    def __init__(self, arg_value: IType | None = None):
        super().__init__(arg_value)

    def generate_internal_opcodes(self, code_generator):
        code_generator.convert_builtin_method_call(NativeContract.StdLib.class_methods['str_len'])



import ast

from boa3.internal.model.expression import IExpression
from boa3.internal.model.method import Method
from boa3.internal.model.type.itype import IType


class Property(IExpression):
    """
    A class used to represent a variable

    :ivar var_type: the type of the variable.
    """

    def __init__(self,
                 getter: Method,
                 setter: Method = None,
                 origin_node: ast.AST | None = None,
                 deprecated: bool = False
                 ):
        super().__init__(origin_node, deprecated)
        self._getter: Method = getter
        self._setter: Method | None = setter

    @property
    def shadowing_name(self) -> str:
        return 'property'

    @property
    def type(self) -> IType:
        return self._getter.type

    @property
    def getter(self) -> Method:
        return self._getter

    @property
    def setter(self) -> Method | None:
        return self._setter

    def __str__(self) -> str:
        return str(self.type)



from typing import Any

from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.builtin.method.strmethod import StrMethod
from boa3.internal.model.type.classes.userclass import UserClass
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class StrClassMethod(StrMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.classes import userclass

        if arg_value is None:
            arg_value = userclass._EMPTY_CLASS

        args: dict[str, Variable] = {
            'object': Variable(arg_value),
        }

        super().__init__(args)

    @property
    def arg_type(self) -> UserClass:
        return self._arg_value.type

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, UserClass):
            if 'object' in self.args and value.is_type_of(self.args['object'].type):
                return self

            return StrClassMethod(value)

        return super().build(value)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.type.type import Type

        # class to dict, then print as json
        code_generator.insert_opcode(Opcode.UNPACK, add_to_stack=[Type.any, Type.int])
        code_generator.remove_stack_top_item()
        code_generator.convert_new_map(Type.dict)

        # add each value to the dict
        for variable_name in self.arg_type.instance_variables:
            code_generator.insert_opcode(Opcode.TUCK, add_to_stack=[Type.any])
            code_generator.convert_literal(variable_name)
            item_address = code_generator.bytecode_size
            code_generator.swap_reverse_stack_items(3, rotate=True)
            code_generator.convert_set_item(item_address, index_inserted_internally=True)

        # print as json
        code_generator.convert_builtin_method_call(Interop.JsonSerialize, is_internal=True)



import ast
from typing import Any

from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.operation.binaryop import BinaryOp
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class SortMethod(IBuiltinMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type

        if not Type.list.is_type_of(arg_value):
            arg_value = Type.list

        identifier = 'sort'
        args: dict[str, Variable] = {
            'self': Variable(arg_value),
            'reverse': Variable(Type.bool)
        }
        # TODO: change this when keyword-only arguments are implemented #2ewewtz
        kwargs = {
            'reverse': Variable(Type.bool)
        }
        reverse_default = ast.parse(str(Type.bool.default_value)).body[0].value

        super().__init__(identifier, args, kwargs=kwargs, defaults=[reverse_default])

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type

        if self._arg_self.type is Type.list:
            return self._identifier

        if Type.list.is_type_of(self._arg_self.type):
            return '-{0}_{1}'.format(self._identifier, self._arg_self.type.identifier)

        return self._identifier

    @property
    def is_supported(self) -> bool:
        from boa3.internal.model.type.primitive.primitivetype import PrimitiveType
        item_type: IType = self._arg_self.type.item_type
        if isinstance(item_type, PrimitiveType):
            return True

        from boa3.internal.model.type.annotation.uniontype import UnionType
        if (isinstance(item_type, UnionType) and
                all(isinstance(union_type, PrimitiveType) for union_type in item_type.union_types)):
            return True

        return False

    @property
    def _arg_self(self) -> Variable:
        return self.args['self']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        # code_generator.remove_stack_top_item()  # ignore key arg
        # reverse, self // reverse -> top

        # start_index = 0
        # i = 1
        # end_address = len(self)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(1)

        # while i <= end_address
        begin_i_while_address = code_generator.convert_begin_while()

        code_generator.duplicate_stack_item(4)  # push array to the stack top
        #   j = i
        code_generator.duplicate_stack_item(2)

        j_while_start_address = code_generator.convert_begin_while()
        #   while j > start_index and arr[j] < arr[j - 1]:
        #       arr[j], arr[j - 1] = arr[j - 1], arr[j]
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.DEC)    # (j - 1), j, arr

        arr_j_minus_one_address = code_generator.bytecode_size
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)  # aux = arr[j - 1]

        # arr[j - 1] / j - 1 / j
        code_generator.swap_reverse_stack_items(3, rotate=True)  # j / arr[j - 1] / j - 1
        arr_j_address = code_generator.bytecode_size
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)  # arr[j]

        # arr[j] / j / arr[j - 1] / j - 1 / arr
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(6)
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_set_item(arr_j_address, index_inserted_internally=True)  # arr[j - 1] = arr[j]

        # j / arr[j - 1] / j - 1 / arr
        code_generator.duplicate_stack_item(4)
        code_generator.swap_reverse_stack_items(3)
        code_generator.convert_set_item(arr_j_minus_one_address, index_inserted_internally=True)  # arr[j] = arr[j - 1]

        j_while_test_address = code_generator.bytecode_size
        # j > start_index
        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.duplicate_stack_top_item()
        if_j_positive = code_generator.convert_begin_if()

        # arr[j] (> if reverse else <) arr[j - 1]
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_get_item(index_inserted_internally=True)
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(4)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.convert_get_item(index_inserted_internally=True)

        # "gt" if reverse else "lt"
        code_generator.duplicate_stack_item(8)
        if_is_reverse = code_generator.convert_begin_if()
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        else_is_reverse = code_generator.convert_begin_else(if_is_reverse, is_internal=True)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)

        code_generator.convert_end_if(else_is_reverse, is_internal=True)

        code_generator.convert_operation(BinaryOp.And, is_internal=True)

        code_generator.convert_end_if(if_j_positive, is_internal=True)

        code_generator.convert_end_while(j_while_start_address, j_while_test_address, is_internal=True)

        code_generator.remove_stack_top_item()  # remove j from the stack
        code_generator.remove_stack_top_item()  # remove array from the stack

        #   i += 1
        code_generator.insert_opcode(Opcode.INC)

        condition_i_while_address = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.convert_end_while(begin_i_while_address, condition_i_while_address, is_internal=True)

        # clean stack
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

    def push_self_first(self) -> bool:
        return self.has_self_argument

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if not isinstance(value, list):
            value = [value]

        from boa3.internal.model.type.type import Type
        if len(value) > 0 and Type.list.is_type_of(value[0]):
            return SortMethod(value[0])

        return self



from boa3.internal.model.builtin.interop.nativecontract import NeoContractMethod
from boa3.internal.model.variable import Variable


class GetCandidateVoteMethod(NeoContractMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType

        identifier = 'get_candidate_vote'
        native_identifier = 'getCandidateVote'
        args: dict[str, Variable] = {
            'pubkey': Variable(ECPointType.build())
        }
        super().__init__(identifier, native_identifier, args,
                         return_type=Type.int)



import ast

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class SumMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'sum'
        args: dict[str, Variable] = {'__iterable': Variable(Type.sequence.build_collection(Type.int)),
                                     '__start': Variable(Type.int)}

        start_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value
        super().__init__(identifier, args, defaults=[start_default], return_type=Type.int)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        # index = 0
        code_generator.convert_literal(0)

        # while index < len(iterable)
        while_start = code_generator.convert_begin_while()

        code_generator.swap_reverse_stack_items(3)
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)
        # sum += iterable[index]
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)
        code_generator.swap_reverse_stack_items(3)
        # index += 1
        code_generator.insert_opcode(Opcode.INC)

        # while condition and end
        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        # remove auxiliary values from stack
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None



import ast

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class Nep17InterfaceMethod(IBuiltinMethod):

    def __init__(self, args: dict[str, Variable], identifier: str,
                 return_type: IType, defaults: list[ast.AST] = None,
                 native_identifier: str = None):
        super().__init__(identifier, args, return_type=return_type, defaults=defaults)
        if native_identifier is None:
            self.native_identifier = identifier
        else:
            self.native_identifier = native_identifier

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo3.contracts import CallFlags

        if len(self.args) != 1:
            code_generator.swap_reverse_stack_items(len(self.args))
            code_generator.swap_reverse_stack_items(len(self.args) - 1)

        code_generator.convert_new_array(len(self.args) - 1)
        code_generator.convert_literal(CallFlags.ALL)
        code_generator.convert_literal(self.native_identifier)
        code_generator.duplicate_stack_item(4)
        code_generator.convert_literal(2)   # Nep17Contract script hash
        code_generator.convert_get_item(index_inserted_internally=True, index_is_positive=True, test_is_negative_index=False)
        code_generator.convert_builtin_method_call(Interop.CallContract, is_internal=True)
        code_generator.remove_stack_item(2)

        if self.return_type is Type.none:
            code_generator.remove_stack_top_item()



import ast

from boa3.internal.model.builtin.method.intmethod import IntMethod
from boa3.internal.model.variable import Variable


class IntIntMethod(IntMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        args: dict[str, Variable] = {
            'value': Variable(Type.int),
        }

        value_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value

        super().__init__(args, [value_default])

    def generate_internal_opcodes(self, code_generator):
        # it is the identity function, so there is no need of including another opcode
        super().generate_internal_opcodes(code_generator)



from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.collection.sequence.tupletype import TupleType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PrintMethod(IBuiltinMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        identifier = 'print'
        if not isinstance(arg_value, IType):
            arg_value = Type.str

        args: dict[str, Variable] = {}
        vararg = ('values', Variable(arg_value))
        super().__init__(identifier, args, return_type=Type.none, vararg=vararg)

    @property
    def _arg_values(self) -> Variable:
        return self._vararg[1]

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_values.type is Type.str:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_values.type._identifier)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 1:
            return False
        if not isinstance(params[0], IExpression):
            return False
        return isinstance(params[0].type, SequenceType)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.builtin.interop.interop import Interop
        from boa3.internal.model.operation.binaryop import BinaryOp

        values_to_print = (self._arg_values.type.value_type
                           if hasattr(self._arg_values.type, 'value_type')
                           else self._arg_values.type)
        # if len(arg) > 0:
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        if_not_empty = code_generator.convert_begin_if()

        #   aux_list = arg.copy().reverse()
        code_generator.convert_copy()
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.SequenceReverse, is_internal=True)

        #   while len(aux_list) > 0
        start_loop = code_generator.convert_begin_while()
        #       _internal_print(aux_list.pop())
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.POPITEM, add_to_stack=[values_to_print], pop_from_stack=True)
        self._generate_print_opcodes(code_generator)
        code_generator.convert_builtin_method_call(Interop.Log, is_internal=True)

        start_loop_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        code_generator.convert_end_while(start_loop, start_loop_condition, is_internal=True)

        code_generator.convert_end_if(if_not_empty, is_internal=True)

        # clear stack
        code_generator.remove_stack_top_item()

    def _generate_print_opcodes(self, code_generator):
        """
        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        pass

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and len(value) == 1:
            value = value[0]
        if isinstance(value, TupleType):
            value = value.value_type
        if type(value) == type(self._arg_values.type):
            return self

        from boa3.internal.model.builtin.method.printbytestringmethod import PrintByteStringMethod
        from boa3.internal.model.type.classes.userclass import UserClass
        from boa3.internal.model.type.type import Type

        if Type.bool.is_type_of(value):
            from boa3.internal.model.builtin.method.printboolmethod import PrintBoolMethod
            return PrintBoolMethod()

        elif Type.int.is_type_of(value):
            from boa3.internal.model.builtin.method.printintmethod import PrintIntMethod
            return PrintIntMethod()

        elif Type.str.is_type_of(value) or Type.bytes.is_type_of(value):
            return PrintByteStringMethod(value)

        elif isinstance(value, UserClass):
            from boa3.internal.model.builtin.method.printclassmethod import PrintClassMethod
            return PrintClassMethod(value)

        elif Type.sequence.is_type_of(value):
            from boa3.internal.model.builtin.method.printsequencemethod import PrintSequenceMethod
            return PrintSequenceMethod(value)

        return PrintByteStringMethod(value)



import ast

from boa3.internal.model.callable import Callable
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.type.type import Type
from boa3.internal.model.variable import Variable


class Event(Callable, IdentifiedSymbol):
    def __init__(self,
                 event_id: str, args: dict[str, Variable] = None,
                 vararg: tuple[str, Variable] | None = None,
                 kwargs: dict[str, Variable] | None = None,
                 defaults: list[ast.AST] = None,
                 origin_node: ast.AST | None = None,
                 deprecated: bool = False
                 ):
        super().__init__(args, vararg, kwargs, defaults, Type.none, True, origin_node, deprecated=deprecated)

        self.name: str = event_id
        self._identifier: str = None

    @property
    def shadowing_name(self) -> str:
        return 'event'

    @property
    def identifier(self) -> str:
        if self._identifier is None:
            # internal identifier should not be the name to avoid nonexistent duplicated symbol ids
            self._identifier = f'-{self.name}-{hex(id(self))}'
        return self._identifier

    @property
    def args_to_generate(self) -> dict[str, Variable]:
        return self.args.copy()

    @property
    def generate_name(self) -> bool:
        return True



import ast
import importlib.util
import os
import sys

from boa3.internal import constants
from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.model import imports
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.type import Type


class ImportAnalyser(IAstAnalyser):

    def __init__(self, import_target: str, root_folder: str,
                 importer_file: str | None = None,
                 import_stack: list[str] = None,
                 already_imported_modules: dict = None,
                 log: bool = False,
                 fail_fast: bool = True,
                 get_entry: bool = False):

        self.can_be_imported: bool = False
        self.is_builtin_import: bool = False
        self.is_namespace_package: bool = False
        self.recursive_import: bool = False
        self._import_identifier: str = import_target

        self._get_from_entry: bool = get_entry and (os.path.isfile(import_target) or os.path.isdir(import_target))

        from boa3.internal.model.imports.package import Package
        self._package: Package | None = None
        from boa3.internal.analyser.analyser import Analyser
        self._imported_files: dict[str, Analyser] = (already_imported_modules
                                                     if isinstance(already_imported_modules, dict)
                                                     else {})
        self._import_stack: list[str] = import_stack if isinstance(import_stack, list) else []
        self.analyser: Analyser = None  # set if the import is successful
        self._submodule_search_locations = None

        if isinstance(root_folder, str):
            if os.path.isfile(root_folder):
                root = os.path.dirname(root_folder)
            elif os.path.isdir(root_folder):
                root = root_folder
            else:
                root = os.path.dirname(importer_file)
        else:
            root = os.path.dirname(importer_file)

        super().__init__(ast.Module(body=[]), root_folder=root, log=log, fail_fast=fail_fast)

        if self._get_from_entry:
            self.path: str = import_target.replace(os.sep, constants.PATH_SEPARATOR)
            self._find_package(import_target, importer_file)
            return

        importer_file_dir = os.path.dirname(importer_file)
        sys.path.insert(0, self.root_folder)
        sys.path.insert(1, importer_file_dir)
        try:
            import_spec = importlib.util.find_spec(import_target)
            module_origin: str = import_spec.origin
        except BaseException:
            return
        finally:
            sys.path.remove(importer_file_dir)
            sys.path.remove(self.root_folder)

        self._importer_file = importer_file
        is_importing_a_module = module_origin is not None

        if not is_importing_a_module:
            self.is_namespace_package = True
            if import_spec.submodule_search_locations is None:
                return

            # submodule_search_locations type doesn't support indexing, so we need to iterate to get the first
            for module_ in import_spec.submodule_search_locations:
                module_origin: str = module_  # + "\\" + names[0] + '.py'
                break

            # if the origin is still not set, there's nothing to import
            if module_origin is None:
                return

        path: list[str] = module_origin.split(os.sep)
        self.filename = path[-1]
        self._submodule_search_locations = import_spec.submodule_search_locations
        self.path: str = module_origin.replace(os.sep, constants.PATH_SEPARATOR)

        self._find_package(module_origin, importer_file)

    @property
    def tree(self) -> ast.AST:
        return self._tree

    @property
    def is_import_deprecated(self) -> bool:
        return self._package.is_deprecated if self._package is not None else False

    def export_symbols(self, identifiers: list[str] = None) -> dict[str, ISymbol]:
        """
        Gets a dictionary that maps each exported symbol with its identifier

        :param identifiers: list of identifiers of the imported symbols
        :return:
        """
        if identifiers is None:
            identifiers = list([name for name, symbol in self.symbols.items() if symbol is not None])
        if not self.can_be_imported or not isinstance(identifiers, (list, str)):
            return {}

        if isinstance(identifiers, str):
            identifiers = [identifiers]

        if constants.IMPORT_WILDCARD in identifiers:
            symbols = self.symbols.copy()
        else:
            symbols = {symbol_id: symbol for symbol_id, symbol in self.symbols.items()
                       if symbol_id in identifiers and symbol is not None}
        return symbols

    def update_external_analysed_files(self, external_files: dict):
        if self.can_be_imported and self.path in self._imported_files and self.path not in external_files:
            external_files[self.path] = self._imported_files[self.path]

        for path, analyser in self._imported_files.items():
            if path not in external_files and analyser.is_analysed:
                external_files[path] = analyser

    def _find_package(self, module_origin: str, origin_file: str | None = None):
        path: list[str] = module_origin.split(os.sep)

        package = imports.compilerbuiltin.get_package(self._import_identifier)
        if hasattr(package, 'symbols'):
            if hasattr(package, 'inner_packages'):
                # when have symbol and packages with the same id, prioritize symbol
                self.symbols: dict[str, ISymbol] = package.inner_packages
                self.symbols.update(package.symbols)
            else:
                self.symbols = package.symbols

            self._package = package
            self.can_be_imported = True
            self.is_builtin_import = True
            return

        def is_boa_package() -> bool:
            common_path = os.path.commonpath([self.path, constants.BOA_PACKAGE_PATH])
            if common_path == constants.BOA_PACKAGE_PATH:
                return True
            if 'boa3' not in path:
                return False

            boa_path = path[path.index('boa3'):]
            return len(boa_path) > 1 and boa_path[1] in ('builtin', 'sc')

        if not is_boa_package():
            # doesn't analyse boa3.builtin packages that aren't included in the imports.compilerbuiltin as an user module
            import re

            inside_python_folder = any(re.search(r'python(\d\.?)*', folder.lower()) for folder in path)
            updated_tree = None

            if not (inside_python_folder and 'lib' in path):
                # check circular imports to avoid recursions inside the compiler
                if self.path in self._import_stack:
                    self.recursive_import = True
                    return

                # only user modules and typing lib imports are implemented
                try:
                    if self.path in self._imported_files:
                        analyser = self._imported_files[self.path]
                    else:
                        from boa3.internal.analyser.analyser import Analyser
                        origin = origin_file.replace(os.sep, constants.PATH_SEPARATOR)
                        files = self._import_stack
                        files.append(origin)
                        if self.is_namespace_package:
                            analyser = Analyser(self.tree, module_origin, self.root_folder, log=self._log,
                                                env=self.analyser.env if hasattr(self.analyser, 'env') else None
                                                )
                            if self._include_inner_packages(analyser):
                                analyser.is_analysed = True
                                self._imported_files[self.path] = analyser
                        else:
                            analyser = Analyser.analyse(module_origin, root=self.root_folder,
                                                        imported_files=self._imported_files,
                                                        import_stack=files,
                                                        log=self._log, fail_fast=True)

                            if self._fail_fast and len(analyser.errors) > 0:
                                raise analyser.errors[0]
                            self._include_inner_packages(analyser)

                        if analyser.is_analysed:
                            self._imported_files[self.path] = analyser

                    # include only imported symbols
                    if analyser.is_analysed:
                        for symbol_id, symbol in analyser.symbol_table.items():
                            if symbol_id not in Type.all_types():
                                if not self._get_from_entry:
                                    symbol.defined_by_entry = False
                                self.symbols[symbol_id] = symbol

                    self.errors.extend(analyser.errors)
                    self.warnings.extend(analyser.warnings)

                    updated_tree = analyser.ast_tree
                    self.analyser = analyser
                    self.can_be_imported = analyser.is_analysed
                except FileNotFoundError:
                    self.can_be_imported = False

                if updated_tree is not None:
                    self._tree = updated_tree

    def _include_inner_packages(self, analyser) -> bool:
        if self.path.endswith('.py') and self.filename != f'{constants.INIT_METHOD_ID}.py':
            return False

        import pkgutil
        from boa3.internal.model.imports.importsymbol import Import
        from boa3.internal.model.imports.package import Package

        inner_packages_have_errors = False
        modules = {}
        for importer, modname, is_pkg in pkgutil.iter_modules(self._submodule_search_locations):
            mod_target = self._import_identifier + constants.ATTRIBUTE_NAME_SEPARATOR + modname
            import_analyser = ImportAnalyser(mod_target, self.root_folder,
                                             importer_file=self._importer_file,
                                             import_stack=self._import_stack,
                                             already_imported_modules=self._imported_files,
                                             log=self._log)

            if not inner_packages_have_errors and not import_analyser.can_be_imported:
                inner_packages_have_errors = True

            imported = Package(identifier=modname,
                               other_symbols=import_analyser.symbols,
                               import_origin=Import(import_analyser.path,
                                                    import_analyser._tree,
                                                    import_analyser))
            modules[modname] = imported

        if len(modules) > 0 and hasattr(analyser, 'symbol_table') and isinstance(analyser.symbol_table, dict):
            analyser.symbol_table.update(modules)

        return not inner_packages_have_errors



from typing import Any

from boa3.internal.model.type.enum.intenumtype import IntEnumType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo3.network.payloads.transactionattributetype import \
    TransactionAttributeType as TransactionAttribute


class TransactionAttributeTypeType(IntEnumType):
    """
    A class used to represent Neo TransactionAttributeType
    """

    def __init__(self):
        super().__init__(TransactionAttribute)
        self._identifier = 'TransactionAttributeType'

    @classmethod
    def build(cls, value: Any) -> IType:
        if cls._is_type_of(value):
            from boa3.internal.model.builtin.interop.interop import Interop
            return Interop.TransactionAttributeType

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (TransactionAttribute, TransactionAttributeTypeType))



from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacemethod import Nep17InterfaceMethod
from boa3.internal.model.builtin.interop.contract import ContractType
from boa3.internal.model.variable import Variable


class Nep17InterfaceBalanceOfMethod(Nep17InterfaceMethod):

    def __init__(self, self_type: ContractType):
        from boa3.internal.model.type.type import Type

        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        args: dict[str, Variable] = {
            'self': Variable(self_type),
            'account': Variable(UInt160Type.build())
        }
        super().__init__(args, 'balance_of', native_identifier='balanceOf', return_type=Type.int)



from boa3.internal.model.builtin.interop.storage.put.istorageputmethod import IStoragePutMethod


class StoragePutDictMethod(IStoragePutMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'put_dict'
        value_type = Type.dict

        super().__init__(identifier, value_type=value_type)

    def generate_serialize_value_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop

        converter = Interop.Serialize
        code_generator.convert_builtin_method_call(converter, is_internal=True)



import ast

from boa3.internal.model.builtin.method.strmethod import StrMethod
from boa3.internal.model.variable import Variable


class StrBytesMethod(StrMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type

        args: dict[str, Variable] = {
            'object': Variable(Type.union.build([Type.bytes, Type.str])),
        }
        object_default = ast.parse("'{0}'".format(Type.str.default_value)).body[0].value

        super().__init__(args, [object_default])

    def generate_internal_opcodes(self, code_generator):
        pass



from boa3.internal.model.builtin.interop.storage.get.istoragegetmethod import IStorageGetMethod


class StorageGetListMethod(IStorageGetMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'get_list'
        value_type = Type.list

        super().__init__(identifier, value_type=value_type)

    def generate_default_value_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type
        # default_value = []
        code_generator.convert_new_empty_array(0, Type.list)

    def generate_deserialize_value_opcodes(self, code_generator):
        from boa3.internal.model.builtin.interop.interop import Interop

        converter = Interop.Deserialize
        code_generator.convert_builtin_method_call(converter, is_internal=True)



from boa3.internal.model.builtin.classmethod.copymethod import CopyMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class CopyDictMethod(CopyMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        super().__init__(arg_value if Type.dict.is_type_of(arg_value) else Type.dict)

    def generate_internal_opcodes(self, code_generator):
        code_generator.insert_opcode(Opcode.UNPACK)
        code_generator.insert_opcode(Opcode.PACKMAP)



from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.smart_contract.testcontract import TestContract


class TestRunnerContract(TestContract):
    def __init__(self, contract_name: str, contract_hex_hash: str):
        self._name: str = contract_name
        self._contract_hash: str = contract_hex_hash

        super().__init__(None, None)
        self._script_hash = self._get_script_hash()

    @property
    def name(self) -> str:
        return self._name

    @property
    def path(self) -> str:
        return super().path

    @path.setter
    def path(self, value: str):
        self._nef_path = value

    def _get_script_hash(self) -> bytes | None:
        try:
            return UInt160.from_string(self._contract_hash[2:]
                                       if self._contract_hash.startswith('0x')
                                       else self._contract_hash
                                       ).to_array()
        except BaseException:
            return None

    def is_valid_identifier(self, item: str) -> bool:
        if super().is_valid_identifier(item):
            return True

        # name is also a valid identifier
        return item == self._name



from boa3.internal.model.builtin.builtincallable import IBuiltinCallable
from boa3.internal.model.builtin.classmethod import *
from boa3.internal.model.builtin.contract import *
from boa3.internal.model.builtin.decorator import *
from boa3.internal.model.builtin.internal import *
from boa3.internal.model.builtin.math import *
from boa3.internal.model.builtin.method import *
from boa3.internal.model.callable import Callable
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.type.collection.sequence.ecpointtype import ECPointType
from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type
from boa3.internal.model.type.itype import IType
from boa3.internal.model.type.neo import *


class Builtin:
    @classmethod
    def get_symbol(cls, symbol_id: str) -> Callable | None:
        for method in cls._python_builtins:
            if isinstance(method, IBuiltinCallable) and method.identifier == symbol_id:
                return method

    @classmethod
    def get_by_self(cls, symbol_id: str, self_type: IType) -> Callable | None:
        for name, method in vars(cls).items():
            if (isinstance(method, IBuiltinMethod)
                    and method.identifier == symbol_id
                    and method.validate_self(self_type)):
                return method

    # builtin method
    Abs = AbsMethod()
    Exit = ExitMethod()
    IsInstance = IsInstanceMethod()
    Len = LenMethod()
    NewEvent = CreateEventMethod()
    Max = MaxIntMethod()
    Min = MinIntMethod()
    Print = PrintMethod()
    ScriptHashMethod_ = ScriptHashMethod()
    StrSplit = StrSplitMethod()
    StrSplitWithoutMaxsplit = StrSplitWithoutMaxsplitMethod()
    Sum = SumMethod()
    ToHexStr = ToHexStrMethod()

    # python builtin class constructor
    Bool = BoolMethod()
    ByteArray = ByteArrayMethod()
    ByteArrayEncoding = ByteArrayEncodingMethod()
    Exception = ExceptionMethod()
    IntByteString = IntByteStringMethod()
    IntInt = IntIntMethod()
    ListBytesString = ListBytesStringMethod()
    ListGeneric = ListGenericMethod()
    ListMapping = ListMappingMethod()
    ListSequence = ListSequenceMethod()
    Range = RangeMethod()
    Reversed = ReversedMethod()
    StrBool = StrBoolMethod()
    StrBytes = StrBytesMethod()
    StrClass = StrClassMethod()
    StrInt = StrIntMethod()
    StrSequence = StrSequenceMethod()
    Super = SuperMethod()

    # python class method
    BytesStringIndex = IndexBytesStringMethod()
    BytesStringIsDigit = IsDigitMethod()
    BytesStringJoin = JoinMethod()
    BytesStringLower = LowerMethod()
    BytesStringStartswith = StartsWithMethod()
    BytesStringStrip = StripMethod()
    BytesStringUpper = UpperMethod()
    BytesStringReplace = ReplaceMethod()
    CountSequenceGeneric = CountSequenceGenericMethod()
    CountSequencePrimitive = CountSequencePrimitiveMethod()
    CountStr = CountStrMethod()
    Copy = CopyListMethod()
    ListSort = SortMethod()
    SequenceAppend = AppendMethod()
    SequenceClear = ClearMethod()
    SequenceExtend = ExtendMethod()
    SequenceIndex = IndexSequenceMethod()
    SequenceInsert = InsertMethod()
    SequencePop = PopSequenceMethod()
    SequenceRemove = RemoveMethod()
    SequenceReverse = ReverseMethod()
    DictKeys = MapKeysMethod()
    DictPop = PopDictMethod()
    DictPopDefault = PopDictDefaultMethod()
    DictValues = MapValuesMethod()

    # custom class methods
    ConvertIntToBytes = IntToBytesMethod()
    ConvertStrToBytes = StrToBytesMethod()
    ConvertToInt = ToIntMethod
    ConvertToStr = ToStrMethod
    ConvertToBool = ToBoolMethod

    # builtin decorator
    ClassMethodDecorator = ClassMethodDecorator()
    InstanceMethodDecorator = InstanceMethodDecorator()
    PropertyDecorator = PropertyDecorator()
    StaticMethodDecorator = StaticMethodDecorator()

    _python_builtins: list[IdentifiedSymbol] = [
        Abs,
        ByteArray,
        ByteArrayEncoding,
        BytesStringIndex,
        BytesStringIsDigit,
        BytesStringJoin,
        BytesStringLower,
        BytesStringStartswith,
        BytesStringStrip,
        BytesStringUpper,
        BytesStringReplace,
        ClassMethodDecorator,
        ConvertToBool,
        ConvertIntToBytes,
        ConvertStrToBytes,
        ConvertToInt,
        ConvertToStr,
        Copy,
        CountSequenceGeneric,
        CountSequencePrimitive,
        CountStr,
        DictKeys,
        DictValues,
        Exception,
        Exit,
        IsInstance,
        Len,
        ListSort,
        Max,
        Min,
        Print,
        PropertyDecorator,
        Range,
        Reversed,
        ScriptHashMethod_,
        SequenceAppend,
        SequenceClear,
        SequenceExtend,
        SequenceIndex,
        SequenceInsert,
        SequencePop,
        SequenceRemove,
        SequenceReverse,
        StaticMethodDecorator,
        StrSplit,
        StrSplitWithoutMaxsplit,
        Sum,
        Super,
    ]

    # boa builtin decorator
    ContractInterface = ContractDecorator()
    ContractMethodDisplayName = DisplayNameDecorator()
    Public = PublicDecorator()

    # boa builtin type
    Event = EventType
    UInt160 = UInt160Type.build()
    UInt256 = UInt256Type.build()
    ECPoint = ECPointType.build()
    NeoAccountState = NeoAccountStateType.build()
    Opcode = OpcodeType.build()
    Address = AddressType.build()
    BlockHash = BlockHashType.build()
    PublicKey = PublicKeyType.build()
    ScriptHashType_ = ScriptHashType.build()
    ScriptHashLittleEndian = ScriptHashLittleEndianType.build()
    TransactionId = TransactionIdType.build()

    # boa events
    Nep11Transfer = Nep11TransferEvent()
    Nep17Transfer = Nep17TransferEvent()

    # boa contract interfaces
    Nep17Contract = Nep17ContractClass()

    # boa smart contract methods
    Abort = AbortMethod()
    Env = EnvProperty.build()

    # region boa builtin modules

    BuiltinMathCeil = DecimalCeilingMethod()
    BuiltinMathFloor = DecimalFloorMethod()

    # endregion

    metadata_fields: dict[str, type | tuple[type, ...]] = {
        'name': str,
        'source': (str, type(None)),
        'supported_standards': list,
        'trusts': list,
        'author': (str, type(None)),
        'email': (str, type(None)),
        'description': (str, type(None)),
        'extras': dict
    }

    _internal_methods = [InnerDeployMethod.instance()
                         ]
    internal_methods = {method.raw_identifier: method for method in _internal_methods}



from boa3.internal.model.builtin.interop.contract import ContractType
from boa3.internal.model.builtin.interop.nativecontract import ContractManagementMethod
from boa3.internal.model.variable import Variable


class GetContractByIdMethod(ContractManagementMethod):

    def __init__(self, contract_type: ContractType):
        from boa3.internal.model.type.type import Type

        identifier = 'get_contract_by_id'
        syscall = 'getContractById'
        args: dict[str, Variable] = {
            'contract_id': Variable(Type.int),
        }

        super().__init__(identifier, syscall, args, return_type=contract_type)



from abc import ABC

from boa3.internal.model.builtin.builtinsymbol import IBuiltinSymbol
from boa3.internal.model.method import Method
from boa3.internal.model.property import Property


class IBuiltinProperty(Property, IBuiltinSymbol, ABC):
    def __init__(self,
                 identifier: str,
                 getter: Method,
                 setter: Method = None,
                 deprecated: bool = False
                 ):
        super().__init__(getter, setter, deprecated=deprecated)
        self._identifier = identifier

    def update_with_analyser(self, analyser):
        super().update_with_analyser(analyser)

        if hasattr(self._getter, 'update_with_analyser'):
            self._getter.update_with_analyser(analyser)

        if hasattr(self._setter, 'update_with_analyser'):
            self._setter.update_with_analyser(analyser)



from typing import Any, Self

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.method import Method
from boa3.internal.model.property import Property
from boa3.internal.model.type.classes.classarraytype import ClassArrayType
from boa3.internal.model.variable import Variable


class BlockType(ClassArrayType):
    """
    A class used to represent Neo Block class
    """

    def __init__(self):
        super().__init__('Block')
        from boa3.internal.model.type.type import Type
        from boa3.internal.model.type.collection.sequence.uint160type import UInt160Type
        from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type

        uint256 = UInt256Type.build()

        self._variables: dict[str, Variable] = {
            'hash': Variable(uint256),
            'version': Variable(Type.int),
            'previous_hash': Variable(uint256),
            'merkle_root': Variable(uint256),
            'timestamp': Variable(Type.int),
            'nonce': Variable(Type.int),
            'index': Variable(Type.int),
            'primary_index': Variable(Type.int),
            'next_consensus': Variable(UInt160Type.build()),
            'transaction_count': Variable(Type.int)
        }
        self._constructor: Method = None

    @property
    def class_variables(self) -> dict[str, Variable]:
        return {}

    @property
    def instance_variables(self) -> dict[str, Variable]:
        return self._variables.copy()

    @property
    def properties(self) -> dict[str, Property]:
        return {}

    @property
    def static_methods(self) -> dict[str, Method]:
        return {}

    @property
    def class_methods(self) -> dict[str, Method]:
        return {}

    @property
    def instance_methods(self) -> dict[str, Method]:
        return {}

    def constructor_method(self) -> Method | None:
        # was having a problem with recursive import
        if self._constructor is None:
            self._constructor: Method = BlockMethod(self)
        return self._constructor

    @classmethod
    def build(cls, value: Any = None) -> Self:
        if value is None or cls._is_type_of(value):
            return _Block

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, BlockType)


_Block = BlockType()


class BlockMethod(IBuiltinMethod):

    def __init__(self, return_type: BlockType):
        identifier = '-Block__init__'
        args: dict[str, Variable] = {}
        super().__init__(identifier, args, return_type=return_type)

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 0

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.neo3.core.types import UInt160, UInt256

        uint160_default = UInt160.zero().to_array()
        uint256_default = UInt256.zero().to_array()

        code_generator.convert_literal(0)  # transaction_count
        code_generator.convert_literal(uint160_default)  # next_consensus
        code_generator.convert_literal(0)  # primary_index
        code_generator.convert_literal(0)  # index
        code_generator.convert_literal(0)  # nonce
        code_generator.convert_literal(0)  # timestamp
        code_generator.convert_literal(uint256_default)  # merkle_root
        code_generator.convert_literal(uint256_default)  # previous_hash
        code_generator.convert_literal(0)  # version
        code_generator.convert_literal(uint256_default)  # hash
        code_generator.convert_new_array(length=len(self.type.symbols), array_type=self.type)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return



__all__ = [
    'GetAttributeFeeMethod',
    'GetExecFeeFactorMethod',
    'GetFeePerByteMethod',
    'GetStoragePriceMethod',
    'IsBlockedMethod',
    'SetAttributeFeeMethod',
    'TransactionAttributeTypeType',
]

from boa3.internal.model.builtin.interop.policy.getattributefeemethod import GetAttributeFeeMethod
from boa3.internal.model.builtin.interop.policy.getexecfeefactormethod import GetExecFeeFactorMethod
from boa3.internal.model.builtin.interop.policy.getfeeperbytemethod import GetFeePerByteMethod
from boa3.internal.model.builtin.interop.policy.getstoragepricemethod import GetStoragePriceMethod
from boa3.internal.model.builtin.interop.policy.isblockedmethod import IsBlockedMethod
from boa3.internal.model.builtin.interop.policy.setattributefeemethod import SetAttributeFeeMethod
from boa3.internal.model.builtin.interop.policy.transactionattributetypetype import TransactionAttributeTypeType



import ast
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class RangeMethod(IBuiltinMethod):

    def __init__(self, values_types: list[IType] = None):
        from boa3.internal.model.type.type import Type
        identifier = 'range'
        if not isinstance(values_types, list) or len(values_types) < 1:
            stop = Variable(Type.none)
            start = Variable(Type.none)
            step = Variable(Type.none)
        else:
            stop = Variable(Type.int)
            start = Variable(Type.int)
            step = Variable(Type.int)

        args: dict[str, Variable] = {
            'stop': stop,
            'start': start,
            'step': step
        }
        start_default = ast.parse("{0}".format(Type.int.default_value)
                                  ).body[0].value
        step_default = ast.parse("1").body[0].value
        super().__init__(identifier, args, defaults=[start_default, step_default], return_type=Type.range)

    @property
    def _arg_stop(self) -> Variable:
        return self.args['stop']

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self._arg_stop.type is Type.none:
            return self._identifier
        return '-{0}_from_{1}'.format(self._identifier, self._arg_stop.type._identifier)

    @property
    def requires_reordering(self) -> bool:
        return True

    def reorder(self, arguments: list):
        if len(arguments) > 1:
            # swap start and stop default positions
            arguments[0], arguments[1] = arguments[1], arguments[0]

    def validate_parameters(self, *params: IExpression | IType) -> bool:
        if len(params) < 1 or len(params) > 3:
            return False

        if any(not isinstance(param, (IExpression, IType)) for param in params):
            return False

        params_type: list[IType] = [param if isinstance(param, IType) else param.type for param in params]
        from boa3.internal.model.type.type import Type
        return all(param is Type.int for param in params_type)

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp

        range_error_msg = 'range() arg 3 must not be zero'

        code_generator.duplicate_stack_item(3)
        code_generator.insert_opcode(Opcode.SIGN)
        # if step == 0
        if_invalid_step = code_generator.convert_begin_if()
        code_generator.change_jump(if_invalid_step, Opcode.JMPIF)
        #   raise error
        code_generator.convert_literal(range_error_msg)
        code_generator.convert_raise_exception()

        code_generator.convert_end_if(if_invalid_step, is_internal=True)

        # aux = []
        code_generator.convert_literal([])
        # new_item = start
        code_generator.swap_reverse_stack_items(4)
        code_generator.swap_reverse_stack_items(2)

        # while new_item is valid
        begin_while = code_generator.convert_begin_while()

        #   aux.append(new_item)
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend, is_internal=True)
        #   new_item += step
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Add, is_internal=True)

        # while condition
        test_while = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(4)
        code_generator.insert_opcode(Opcode.SIGN)
        code_generator.convert_literal(0)

        # if step <= 0 -> condition is GT
        if_condition = code_generator.convert_begin_if()
        code_generator.change_jump(if_condition, Opcode.JMPGT)  # jmp to negation of LE
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        # else -> condition is LT
        else_condition = code_generator.convert_begin_else(if_condition, is_internal=True)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.convert_end_if(else_condition, is_internal=True)

        code_generator.convert_end_while(begin_while, test_while, is_internal=True)

        # clear stack
        for _ in self.args:
            code_generator.remove_stack_top_item()

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, list) and self.validate_parameters(*value):
            return RangeMethod(value)
        return super().build(value)



__all__ = ['InnerDeployMethod',
           'EnvProperty',
           ]

from boa3.internal.model.builtin.internal.getenvmethod import Env as EnvProperty
from boa3.internal.model.builtin.internal.innerdeploymethod import InnerDeployMethod



from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.builtin.interop.runtime.triggertypetype import TriggerTypeType
from boa3.internal.model.variable import Variable


class GetTriggerMethod(InteropMethod):

    def __init__(self, trigger_type: TriggerTypeType):
        identifier = 'get_trigger'
        syscall = 'System.Runtime.GetTrigger'
        args: dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=trigger_type)



from typing import Any

from boa3.internal.model.type.enum.intflagtype import IntFlagType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo3.contracts import CallFlags


class CallFlagsType(IntFlagType):
    """
    A class used to represent Neo interop CallFlags type
    """

    def __init__(self):
        super().__init__(CallFlags)
        self._identifier = 'CallFlags'

    @property
    def default_value(self) -> Any:
        return CallFlags.ALL

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if value is None or cls._is_type_of(value):
            return _CallFlags

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (CallFlags, CallFlagsType))

_CallFlags = CallFlagsType()



import ast
from typing import Self

from boa3.internal.model.expression import IExpression
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.itype import IType


class GeneratorData:

    def __init__(self, origin_node: ast.AST | None,
                 symbol_id: str | None = None,
                 symbol: ISymbol | None = None,
                 result_type: IType | None = None,
                 index: int | None = None,
                 origin_object_type: IType | None = None,
                 already_generated: bool = False):

        if symbol_id is None and isinstance(symbol, IdentifiedSymbol):
            symbol_id = symbol.identifier
        if result_type is None and isinstance(symbol, IExpression):
            result_type = symbol.type

        self.node: ast.AST | None = origin_node
        self.symbol: ISymbol | None = symbol
        self.symbol_id: str | None = symbol_id
        self.type: IType | None = result_type
        self.index: int | None = index
        self.origin_object_type: IType | None = origin_object_type
        self.already_generated: bool = already_generated

    def copy(self, new_origin: ast.AST | None = None) -> Self:
        return GeneratorData(
            origin_node=new_origin if isinstance(new_origin, ast.AST) else self.node,
            symbol_id=self.symbol_id,
            symbol=self.symbol,
            result_type=self.type,
            index=self.index,
            origin_object_type=self.origin_object_type,
            already_generated=self.already_generated
        )



from enum import IntEnum


class NamedCurveHash(IntEnum):
    """
    Represents the named curve used in ECDSA.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/foundation/Cryptography/encryption_algorithm#ecdsa-signing>`__
    to learn more about ECDSA signing.
    """

    SECP256K1SHA256 = 22
    """
    The secp256k1 curve and SHA256 hash algorithm.

    :meta hide-value: 
    """

    SECP256R1SHA256 = 23
    """
    The secp256r1 curve, which known as prime256v1 or nistP-256, and SHA256 hash algorithm.

    :meta hide-value:
    """

    SECP256K1KECCAK256 = 122
    """
    The secp256k1 curve and Keccak256 hash algorithm.

    :meta hide-value:
    """

    SECP256R1KECCAK256 = 123
    """
    The secp256r1 curve, which known as prime256v1 or nistP-256, and Keccak256 hash algorithm.

    :meta hide-value:
    """



import ast
from abc import abstractmethod

from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.itype import IType


class IExpression(ISymbol):
    """
    An interface used to represent expressions
    """

    def __init__(self, origin_node: ast.AST | None = None, deprecated: bool = False):
        self._origin_node = origin_node
        self._deprecated = deprecated

    @property
    @abstractmethod
    def type(self) -> IType:
        """
        Gets the type of the evaluated expression

        :return: the resulting type when the expression is evaluated
        """
        pass

    @property
    def is_deprecated(self) -> bool:
        return self._deprecated

    def deprecate(self, new_location: str = None):
        self._deprecated = True

    @property
    def origin(self) -> ast.AST:
        """
        Returns the method origin ast node.

        :return: the ast node that describes this method. None if it is not from a ast.
        """
        return self._origin_node



from boa3.internal.model.builtin.interop.blockchain.transactiontype import TransactionType
from boa3.internal.model.builtin.interop.nativecontract import LedgerMethod
from boa3.internal.model.variable import Variable


class GetTransactionMethod(LedgerMethod):

    def __init__(self, transaction_type: TransactionType):
        from boa3.internal.model.type.collection.sequence.uint256type import UInt256Type
        from boa3.internal.model.type.type import Type

        identifier = 'get_transaction'
        syscall = 'getTransaction'
        args: dict[str, Variable] = {'hash_': Variable(UInt256Type.build())}
        super().__init__(identifier, syscall, args, return_type=Type.optional.build(transaction_type))



from enum import Enum

from boa3.internal.model.builtin.builtin import Builtin
from boa3.internal.model.builtin.compile_time import NeoMetadataType
from boa3.internal.model.builtin.interop.contract.contractmanifest import *
from boa3.internal.model.builtin.interop.interop import Interop
from boa3.internal.model.builtin.native.nativecontract import NativeContract
from boa3.internal.model.imports.package import Package
from boa3.internal.model.type.math import Math


class BoaSCPackage(str, Enum):
    Contracts = 'contracts'
    Compiletime = 'compiletime'
    Math = 'math'
    Runtime = 'runtime'
    Storage = 'storage'
    Types = 'types'
    Utils = 'utils'


class ContractImports:

    @classmethod
    def package_symbols(cls, package: str = None) -> Package | None:
        if package in BoaSCPackage.__members__.values():
            return cls._pkg_tree[package]

    # region Contracts
    ContractManagementModule = Package(
        identifier=NativeContract.ContractManagement.identifier.lower(),
        types=[
            NativeContract.ContractManagement,
        ]
    )

    CryptoLibModule = Package(
        identifier=NativeContract.CryptoLib.identifier.lower(),
        types=[
            NativeContract.CryptoLib,
        ]
    )

    GasModule = Package(
        identifier=NativeContract.GAS.identifier.lower(),
        types=[
            NativeContract.GAS
        ]
    )

    LedgerModule = Package(
        identifier=NativeContract.Ledger.identifier.lower(),
        types=[
            NativeContract.Ledger
        ]
    )

    NeoModule = Package(
        identifier=NativeContract.NEO.identifier.lower(),
        types=[
            NativeContract.NEO
        ]
    )

    OracleModule = Package(
        identifier=NativeContract.Oracle.identifier.lower(),
        types=[
            NativeContract.Oracle
        ]
    )

    PolicyModule = Package(
        identifier=NativeContract.Policy.identifier.lower(),
        types=[
            NativeContract.Policy
        ]
    )

    RoleManagementModule = Package(
        identifier=NativeContract.RoleManagement.identifier.lower(),
        types=[
            NativeContract.RoleManagement,
        ]
    )

    StdLibModule = Package(
        identifier=NativeContract.StdLib.identifier.lower(),
        types=[
            NativeContract.StdLib
        ]
    )

    ContractsPackage = Package(
        identifier=BoaSCPackage.Contracts,
        types=[
            NativeContract.ContractManagement,
            NativeContract.CryptoLib,
            NativeContract.GAS,
            NativeContract.Ledger,
            NativeContract.NEO,
            NativeContract.Oracle,
            NativeContract.Policy,
            NativeContract.RoleManagement,
            NativeContract.StdLib,
        ],
        packages=[
            ContractManagementModule,
            CryptoLibModule,
            GasModule,
            LedgerModule,
            NeoModule,
            OracleModule,
            PolicyModule,
            RoleManagementModule,
            StdLibModule,
        ],
    )
    # endregion

    # region Compiletime
    CompiletimePackage = Package(
        identifier=BoaSCPackage.Compiletime,
        types=[
            NeoMetadataType,
        ],
        methods=[
            Builtin.Public,
            Builtin.ContractInterface,
            Builtin.ContractMethodDisplayName,
        ]
    )
    # endregion

    # region Math
    MathPackage = Package(
        identifier=BoaSCPackage.Math,
        methods=[
            Math.Sqrt,
            Builtin.BuiltinMathCeil,
            Builtin.BuiltinMathFloor,
        ],
    )
    # endregion

    # region Runtime
    RuntimePackage = Package(
        identifier=BoaSCPackage.Runtime,
        properties=[
            Interop.AddressVersion,
            Interop.BlockTime,
            Interop.CallingScriptHash,
            Interop.ExecutingScriptHash,
            Interop.GasLeft,
            Interop.Platform,
            Interop.InvocationCounter,
            Interop.EntryScriptHash,
            Interop.ScriptContainer,
            Interop.GetCurrentSigners
        ],
        methods=[
            Interop.BurnGas,
            Interop.CheckWitness,
            Interop.GetNetwork,
            Interop.GetNotifications,
            Interop.GetRandom,
            Interop.GetTrigger,
            Interop.LoadScript,
            Interop.Log,
            Interop.Notify
        ],
    )
    # endregion

    # region Storage
    StorageContextModule = Package(
        identifier=Interop.StorageContextType.identifier.lower(),
        types=[
            Interop.StorageContextType
        ]
    )

    StorageMapModule = Package(
        identifier=Interop.StorageMapType.identifier.lower(),
        types=[
            Interop.StorageMapType
        ]
    )

    StoragePackage = Package(
        identifier=BoaSCPackage.Storage,
        types=[
            Interop.StorageContextType,
            Interop.StorageMapType,
        ],
        methods=[
            Interop.StorageDelete,
            Interop.StorageFind,
            Interop.StorageGet,
            Interop.StorageGetInt,
            Interop.StorageGetBool,
            Interop.StorageGetStr,
            Interop.StorageGetList,
            Interop.StorageGetDict,
            Interop.StorageGetObject,
            Interop.StorageGetUInt160,
            Interop.StorageGetUInt256,
            Interop.StorageGetECPoint,
            Interop.StorageGetContext,
            Interop.StorageGetCheck,
            Interop.StorageGetCheckInt,
            Interop.StorageGetCheckBool,
            Interop.StorageGetCheckStr,
            Interop.StorageGetCheckList,
            Interop.StorageGetCheckDict,
            Interop.StorageGetCheckObject,
            Interop.StorageGetCheckUInt160,
            Interop.StorageGetCheckUInt256,
            Interop.StorageGetCheckECPoint,
            Interop.StorageGetReadOnlyContext,
            Interop.StoragePut,
            Interop.StoragePutInt,
            Interop.StoragePutBool,
            Interop.StoragePutStr,
            Interop.StoragePutList,
            Interop.StoragePutDict,
            Interop.StoragePutObject,
            Interop.StoragePutUInt160,
            Interop.StoragePutUInt256,
            Interop.StoragePutECPoint,
        ],
        packages=[
            StorageContextModule,
            StorageMapModule
        ]
    )
    # endregion

    # region Types
    TypesPackage = Package(
        identifier=BoaSCPackage.Types,
        types=[
            Builtin.UInt160,
            Builtin.UInt256,
            Builtin.ECPoint,
            Builtin.Event,
            Builtin.Address,
            Builtin.BlockHash,
            Builtin.PublicKey,
            Builtin.ScriptHashType_,
            Builtin.ScriptHashLittleEndian,
            Builtin.TransactionId,
            Interop.BlockType,
            Interop.SignerType,
            Interop.WitnessConditionType,
            Interop.WitnessCondition,
            Interop.WitnessRuleAction,
            Interop.WitnessRule,
            Interop.WitnessScope,
            Interop.TransactionType,
            Interop.ContractType,
            Interop.ContractManifestType,
            ContractAbiType.build(),
            ContractEventDescriptorType.build(),
            ContractGroupType.build(),
            ContractMethodDescriptorType.build(),
            ContractParameterDefinitionType.build(),
            ContractParameterTypeType.build(),
            ContractPermissionDescriptorType.build(),
            ContractPermissionType.build(),
            Builtin.Nep17Contract,
            Builtin.Opcode,
            Interop.FindOptionsType,
            Interop.NamedCurveHashType,
            Interop.Bls12381Type,
            Interop.RoleType,
            Interop.NotificationType,
            Interop.TriggerType,
            Interop.VMStateType,
            Interop.CallFlagsType,
            Interop.OracleResponseCode,
            Builtin.NeoAccountState,
            Interop.TransactionAttributeType,
        ],
    )
    # endregion

    # region Utils
    IteratorModule = Package(
        identifier=Interop.Iterator._identifier.lower(),
        types=[
            Interop.Iterator
        ]
    )

    UtilsPackage = Package(
        identifier=BoaSCPackage.Utils,
        methods=[
            Builtin.Env,
            Builtin.NewEvent,
            Builtin.Nep11Transfer,
            Builtin.Nep17Transfer,
            Builtin.ToHexStr,
            Builtin.ScriptHashMethod_,
            Interop.CallContract,
            Interop.GetCallFlags,
            Builtin.Abort,
            Interop.CheckSig,
            Interop.CheckMultisig,
            Interop.CreateStandardAccount,
            Interop.CreateMultisigAccount,
            Builtin.ConvertToBool,
            Builtin.ConvertIntToBytes,
            Builtin.ConvertStrToBytes,
            Builtin.ConvertToInt,
            Builtin.ConvertToStr,
            Interop.Hash160,
            Interop.Hash256,
        ],
        types=[
            Interop.Iterator
        ],
        packages=[
            IteratorModule
        ],
    )
    # endregion

    _pkg_tree: dict[BoaSCPackage, Package] = {
        pkg.identifier: pkg for pkg in [
            ContractsPackage,
            CompiletimePackage,
            MathPackage,
            RuntimePackage,
            StoragePackage,
            TypesPackage,
            UtilsPackage,
        ]
    }



import ast
from collections.abc import Callable
from typing import Any

from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.exception import CompilerError
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.builtin.method.isinstancemethod import IsInstanceMethod
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.itype import IType


class BuiltinFunctionCallAnalyser(IAstAnalyser):
    def __init__(self, origin: IAstAnalyser, call: ast.Call, method_id: str, builtin_method: IBuiltinMethod,
                 log: bool, fail_fast: bool = True):
        self._method: IBuiltinMethod = builtin_method
        self.method_id: str = method_id
        super().__init__(call, root_folder=origin.root_folder, log=log, fail_fast=fail_fast)

        self._origin: IAstAnalyser = origin

        # all methods validators must be (IBuiltinMethod, list[IType]) -> None
        self._methods_validators: dict[type[IBuiltinMethod],
                                       Callable[[IBuiltinMethod, list[IType]], None]] = {
            IsInstanceMethod: self._validate_IsInstanceMethod
        }

    @property
    def method(self) -> IBuiltinMethod:
        return self._method

    @property
    def call(self) -> ast.Call:
        return self._tree

    def get_symbol(self, symbol_id: str,
                   is_internal: bool = False,
                   check_raw_id: bool = False) -> ISymbol | None:
        return self._origin.get_symbol(symbol_id, is_internal)

    def get_symbol_from_node(self, node: ast.AST) -> ISymbol | None:
        if isinstance(node, ast.Name):
            return self.get_symbol(node.id)

        if isinstance(node, ast.Attribute):
            value = self.get_symbol_from_node(node.value)
            if hasattr(value, 'symbols') and node.attr in value.symbols:
                return value.symbols[node.attr]

        return None

    def get_type(self, value: Any, use_metadata: bool = False) -> IType:
        return self._origin.get_type(value, use_metadata)

    def validate(self) -> bool:
        """
        Validates the method arguments.

        :return: whether the method have specific validation
        """
        if type(self.method) in self._methods_validators:
            if self.method_id == self.method.raw_identifier:
                # if the identifiers are diffrente, this method was validated already
                validator = self._methods_validators[type(self.method)]
                args = [self.get_type(param) for param in self.call.args]
                validator(self.method, args)
            return True
        return False

    def _validate_IsInstanceMethod(self, method: IsInstanceMethod, args_types: list[IType]):
        """
        Validates the arguments for `isinstance` method

        :param method: instance of the builtin method
        :param args_types: types of the arguments
        """
        last_arg = self.call.args[-1]
        if isinstance(last_arg, ast.Tuple) and all(isinstance(tpe, (ast.Attribute, ast.Name))
                                                   for tpe in last_arg.elts):
            if len(last_arg.elts) == 1:
                # if the types tuple has only one type, remove it from inside the tuple
                last_arg = self.call.args[-1] = last_arg.elts[-1]
                args_types[-1] = args_types[-1].value_type
            elif len(last_arg.elts) > 1:
                # if there are more than one type, updates information in the instance of the method
                types: list[IType] = [self.get_symbol_from_node(name) for name in last_arg.elts]
                method.set_instance_type(types)
                self.call.args[-1] = last_arg.elts[-1]
                return

        from boa3.internal.model.type.annotation.metatype import MetaType
        from boa3.internal.model.type.type import Type
        is_ast_valid = (isinstance(last_arg, ast.Name)
                        or (isinstance(last_arg, ast.NameConstant) and args_types[-1] is Type.none))

        is_id_valid = (hasattr(last_arg, 'id')
                       and last_arg.id != args_types[-1].identifier
                       and last_arg.id != args_types[-1].raw_identifier
                       and isinstance(self.get_type(last_arg, use_metadata=True), MetaType))

        if not is_ast_valid or is_id_valid:
            # if the value is not the identifier of a type
            self._log_error(
                CompilerError.MismatchedTypes(
                    last_arg.lineno, last_arg.col_offset,
                    expected_type_id=type.__name__,
                    actual_type_id=args_types[-1].identifier
                ))



from boa3.internal.model.builtin.method.minmethod import MinMethod
from boa3.internal.model.type.itype import IType


class MinByteStringMethod(MinMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        is_valid_type = Type.str.is_type_of(arg_value) or Type.bytes.is_type_of(arg_value)
        super().__init__(arg_value if is_valid_type else Type.str)

    def _compare_values(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        # if (str1 != str2):
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        if_init_are_not_equal = code_generator.convert_begin_if()
        code_generator.change_jump(if_init_are_not_equal, Opcode.JMPEQ)

        #   min_len_str = min((len(str1), len(str2))
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        self.generate_internal_opcodes(code_generator)
        #   index = 0
        code_generator.convert_literal(0)

        #   while(index <= min_len_str):
        start_while = code_generator.convert_begin_while()

        #       value1 = str1[index]
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)
        code_generator.duplicate_stack_item(2)
        #       value2 = str2[index]
        code_generator.duplicate_stack_item(5)
        code_generator.swap_reverse_stack_items(2)
        code_generator.convert_get_item(index_inserted_internally=True, test_is_negative_index=False)

        #       if (value2 == value1):
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        if_values_not_equal = code_generator.convert_begin_if()
        code_generator.change_jump(if_values_not_equal, Opcode.JMPNE)

        #           index += 1
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.insert_opcode(Opcode.INC)

        #   while condition and end
        condition_address = code_generator.bytecode_size
        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)
        code_generator.convert_end_while(start_while, condition_address, is_internal=True)

        #   if (index > min_len_str):
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        #       is_str2_min = len(str2) < len(str1)
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)

        #       elif (value1 != value2):
        else_values_are_equal = code_generator.convert_begin_else(if_values_not_equal)
        #           is_str2_min = value1 > value2
        code_generator.convert_operation(BinaryOp.Gt, is_internal=True)
        code_generator.remove_stack_item(2)
        code_generator.remove_stack_item(2)
        code_generator.convert_end_if(else_values_are_equal)

        # elif (str1 == str2):
        else_init_are_equal = code_generator.convert_begin_else(if_init_are_not_equal)
        #   is_str2_min = True
        code_generator.convert_literal(1)
        code_generator.convert_end_if(else_init_are_equal, is_internal=True)

        # if (is_str2_min):
        if_str2_is_min = code_generator.convert_begin_if()
        code_generator.change_jump(if_str2_is_min, Opcode.JMPIF)
        #   return str2
        code_generator.remove_stack_top_item()
        # elif (not is_str2_min):
        else_str1_is_min = code_generator.convert_begin_else(if_str2_is_min, is_internal=True)
        code_generator.remove_stack_item(2)
        #   return str1
        code_generator.convert_end_if(else_str1_is_min)



import ast
from typing import Any

from boa3.internal.model import set_internal_call
from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.primitive.bytestype import BytesType
from boa3.internal.model.variable import Variable
from boa3.internal.model.type.type import Type
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ToIntMethod(IBuiltinMethod):
    def __init__(self):
        identifier = 'to_int'

        args: dict[str, Variable] = {
            'value': Variable(Type.bytes),
            'big_endian': Variable(Type.bool),
            'signed': Variable(Type.bool),
        }
        big_endian_default = set_internal_call(ast.parse("{0}".format(False)
                                                            ).body[0].value)
        signed_default = set_internal_call(ast.parse("{0}".format(True)
                                                     ).body[0].value)

        super().__init__(identifier, args, defaults=[big_endian_default, signed_default],
                         return_type=Type.int)

    @property
    def generation_order(self) -> list[int]:
        value_index = list(self.args).index('value')
        big_endian_index = list(self.args).index('big_endian')
        signed_index = list(self.args).index('signed')

        return [signed_index, value_index, big_endian_index]

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        # stack: signed, value, big_endian

        # if big_endian:
        if_is_big_endian = code_generator.convert_begin_if()
        #   value = value[::-1]
        code_generator.convert_array_negative_stride()

        code_generator.convert_end_if(if_is_big_endian, is_internal=True)
        # stack: signed, value

        code_generator.swap_reverse_stack_items(2)
        # if signed:
        if_is_signed = code_generator.convert_begin_if()

        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_literal(0b10000000)
        code_generator.convert_operation(BinaryOp.BitAnd)

        if_last_byte_not_begins_with_1 = code_generator.convert_begin_if()
        code_generator.change_jump(if_last_byte_not_begins_with_1, Opcode.JMPIF)

        else_is_not_signed = code_generator.convert_begin_else(if_is_signed, is_internal=True)
        #  value = value + b'\x00'
        code_generator.convert_literal(b'\x00')
        code_generator.convert_operation(BinaryOp.Concat)

        code_generator.convert_end_if(if_last_byte_not_begins_with_1, is_internal=True)
        code_generator.convert_end_if(else_is_not_signed, is_internal=True)
        # stack: value

        self.generate_internal_opcodes(code_generator)

    def generate_internal_opcodes(self, code_generator):
        code_generator.convert_cast(Type.int, is_internal=True)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def build(self, value: Any) -> IBuiltinMethod:
        if type(value) == type(self.args['value'].type):
            return self
        if isinstance(value, BytesType):
            return ToIntMethod()
        return super().build(value)


ToInt = ToIntMethod()



import enum

from boa3.internal.model.callable import Callable
from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.imports.package import Package
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.annotation.metatype import metaType
from boa3.internal.model.type.type import Type
from boa3.internal.model.type.typingmethod.casttypemethod import CastTypeMethod


class TypingPackage(str, enum.Enum):
    Collections = 'collections'
    Typing = 'typing'


class TypeUtils:
    @classmethod
    def all_functions(cls) -> dict[str, Callable]:
        from boa3.internal.model.builtin.builtincallable import IBuiltinCallable
        return {tpe._identifier: tpe for tpe in vars(cls).values() if isinstance(tpe, IBuiltinCallable)}

    @classmethod
    def package_symbols(cls, package: str = None) -> Package | None:
        if package in TypingPackage.__members__.values():
            if TypingPackage.Typing not in cls._pkg_tree:
                cls._pkg_tree[TypingPackage.Typing] = Package(
                    identifier=TypingPackage.Typing,
                    other_symbols=cls.get_types_from_typing_lib()
                )
            return cls._pkg_tree[package]

    @classmethod
    def get_types_from_typing_lib(cls) -> dict[str, ISymbol]:
        import typing
        from types import FunctionType

        type_symbols: dict[str, ISymbol] = {}
        all_types: list[str] = typing.__all__

        deprecated_builtins = [
            'List', 'Dict', 'Tuple'
        ]
        deprecated_collections = list(cls._pkg_tree[TypingPackage.Collections].inner_packages['abc'].symbols)

        for t_id in all_types:
            attr = getattr(typing, t_id)
            type_symbol = None
            if not isinstance(attr, FunctionType):
                type_id: str = t_id if t_id in Type.all_types() else t_id.lower()
                if type_id in Type.all_types():
                    type_symbol = Type.all_types()[type_id]
            else:
                function_id: str = t_id if t_id in cls.all_functions() else t_id.lower()
                if function_id in cls.all_functions():
                    type_symbol = cls.all_functions()[function_id]

            if isinstance(type_symbol, ISymbol):
                if t_id in deprecated_builtins:
                    type_symbol = type_symbol.clone()
                    type_symbol.deprecate(t_id.lower())
                elif t_id in deprecated_collections:
                    type_symbol = type_symbol.clone()
                    type_symbol.deprecate(f'{TypingPackage.Collections.value}.abc.{t_id}')

                type_symbols[t_id] = type_symbol

        return type_symbols

    # type for internal validation
    type = metaType

    # Annotation function utils
    cast = CastTypeMethod()

    _internal_validation_symbols: list[IdentifiedSymbol] = [type
                                                            ]

    symbols_for_internal_validation = {symbol.identifier: symbol
                                       for symbol in _internal_validation_symbols}

    _pkg_tree: dict[TypingPackage, Package] = {
        TypingPackage.Collections: Package.create_package(
            package_id=f'{TypingPackage.Collections.value}.abc',
            symbols={
                'Collection': Type.collection,
                'Mapping': Type.mapping,
                'MutableSequence': Type.mutableSequence,
                'Sequence': Type.sequence,
            }
        )
    }



from collections.abc import Sequence
from typing import Any

from boa3.internal.neo.vm.type.AbiType import AbiType


def value_to_parameter(value: Any) -> Any:
    if isinstance(value, (int, str)):
        return value

    if isinstance(value, (bytes, bytearray)):
        return bytes_to_hex(value)

    if isinstance(value, Sequence):
        return [value_to_parameter(x) for x in value]

    if isinstance(value, dict):
        return {
            'type': AbiType.Map.value,
            'value': [{'key': value_to_parameter(key),
                       'value': value_to_parameter(value)
                       } for key, value in value.items()
                      ]
        }

    if value is not None:
        raise TypeError(f"Unsupported type: '{type(value)}'")

    return None


def bytes_to_hex(data: bytes) -> str:
    return '0x' + data.hex()


def handle_return_error(json_result: str) -> dict:
    import json

    index = _get_first_json_closure(json_result)
    result = json.loads(json_result[:index])
    if 'stack' not in result:
        return result

    while index < len(json_result):
        next_index = _get_first_json_closure(json_result, starting_index=index)
        if next_index > index:
            new_object = json.loads(json_result[index:next_index])
            result['stack'].append(new_object)
        else:
            break
        index = next_index

    return result


def _get_first_json_closure(json_result: str, starting_index: int = 0):
    if starting_index < 0:
        starting_index = 0

    open_object = []
    open_array = []

    if len(json_result) == starting_index or json_result[starting_index] != '{':
        return starting_index

    open_object.append(starting_index)
    index = starting_index + 1
    while index < len(json_result) and len(open_object):
        char = json_result[index]
        if char == '{':
            open_object.append(index)
        elif char == '}':
            open_object.pop()
        elif char == '[':
            open_array.append(index)
        elif char == ']':
            open_array.pop()

        index += 1

    return index



from boa3_test.test_drive.neoxp.command.neoexpresscommand.neoexpresscommand import NeoExpressCommand

__all__ = ['CreateCommand']


class CreateCommand(NeoExpressCommand):
    def __init__(self, config_output: str | None = None,
                 node_count: int = None,
                 address_version: int = None,
                 force: bool = False):

        if node_count not in [1, 4, 7]:  # allowed values for node count
            node_count = None

        self.config_output = config_output
        self.node_count = node_count
        self.address_version = address_version
        self.force = force

        command_id = 'create'
        args = []
        if isinstance(config_output, str):
            args.append(config_output)

        super().__init__(command_id, args)

    def _get_options(self) -> dict[str, str]:
        options = super()._get_options()

        if isinstance(self.address_version, int):
            options['--count'] = str(self.address_version)
        if isinstance(self.address_version, int):
            options['--address-version'] = str(self.address_version)
        if self.force:
            options['--force'] = ''

        return options



from typing import Any

from boa3.internal.neo3.core.types import UInt160
from boa3_test.test_drive.model.network.payloads.witnessscope import WitnessScope
from boa3_test.test_drive.model.wallet.account import Account
from boa3_test.test_drive.neoxp.command import utils
from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.icontractcommand import IContractCommand


class ContractRunCommand(IContractCommand):
    def __init__(self, contract: str | UInt160, method: str, *args: Any,
                 account: Account = None,
                 witness_scope: WitnessScope = WitnessScope.CalledByEntry,
                 results_only: bool = False,
                 additional_gas: int = 0,
                 wallet_password: str = None,
                 trace: bool = False,
                 json_output: bool = False,
                 neo_express_data_file: str = None):

        self.contract: UInt160 = contract
        self.method = method
        self.args = args

        self.account = account
        self.witness_scope = witness_scope
        self.password = wallet_password

        self.results = True if not isinstance(account, Account) else results_only
        self.additional_gas = additional_gas if additional_gas > 0 else 0
        self.input = neo_express_data_file
        self.trace = trace
        self.json_output = json_output

        arguments = [str(contract), method]
        arguments.extend(args)

        super().__init__('run', arguments)

    def _get_options(self) -> dict[str, str]:
        options = super()._get_options()

        if isinstance(self.account, Account):
            options['--account'] = self.account.get_identifier()
        if isinstance(self.witness_scope, WitnessScope):
            options['--witness-scope'] = self.witness_scope.name
        if self.results:
            options['--results'] = ''
        if self.additional_gas > 0:
            options['--gas'] = utils.stringify_asset_quantity(self.additional_gas, utils.GAS_DECIMALS)
        if isinstance(self.password, str):
            options['--password'] = self.password
        if self.trace:
            options['--trace'] = ''
        if self.json_output:
            options['--json'] = ''

        return options



from boa3_test.test_drive.neoxp.command.neoexpresscommand.contract.icontractcommand import IContractCommand


class ContractListCommand(IContractCommand):
    def __init__(self, json_output: bool = False,
                 neo_express_data_file: str = None):

        self.input = neo_express_data_file
        self.json_output = json_output

        super().__init__('list')

    def _get_options(self) -> dict[str, str]:
        options = super()._get_options()

        if self.json_output:
            options['--json'] = ''

        return options



from boa3.internal.model.builtin.classmethod.countmethod import CountMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable


class CountSequenceMethod(CountMethod):

    def __init__(self, sequence_type: SequenceType | None = None, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type
        if not isinstance(sequence_type, SequenceType):
            sequence_type = Type.sequence
        if not isinstance(arg_value, IType):
            arg_value = Type.any

        args: dict[str, Variable] = {
            'self': Variable(sequence_type),
            'value': Variable(arg_value)
        }

        super().__init__(args)

    def validate_parameters(self, *params: IExpression) -> bool:
        if len(params) != 2:
            return False
        if not all(isinstance(param, IExpression) for param in params):
            return False

        from boa3.internal.model.type.itype import IType
        sequence_type: IType = params[0].type

        if not isinstance(sequence_type, SequenceType):
            return False
        from boa3.internal.model.type.collection.sequence.mutable.listtype import ListType
        from boa3.internal.model.type.collection.sequence.tupletype import TupleType
        from boa3.internal.model.type.collection.sequence.rangetype import RangeType
        if not isinstance(sequence_type, (ListType, TupleType, RangeType)):
            return False
        return True

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.neo.vm.opcode.Opcode import Opcode

        # aux = self.copy()
        code_generator.convert_copy()

        # count = 0
        code_generator.convert_literal(0)

        # index = len(sequence) - 1
        code_generator.duplicate_stack_item(2)
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        code_generator.insert_opcode(Opcode.DEC)

        # while index >= 0:
        while_start = code_generator.convert_begin_while()

        jmps_to_inc, jmps_to_condition = self._generic_verification(code_generator)

        #   element = sequence[index]
        code_generator.duplicate_stack_item(3)
        code_generator.duplicate_stack_item(2)
        code_generator.convert_get_item(index_inserted_internally=True)

        #   if element == value:
        code_generator.duplicate_stack_item(5)
        is_equal = code_generator.convert_begin_if()
        code_generator.change_jump(is_equal, Opcode.JMPNE)

        for address in jmps_to_inc:
            code_generator.convert_end_if(address, is_internal=True)

        #       count += 1
        code_generator.swap_reverse_stack_items(2)
        code_generator.insert_opcode(Opcode.INC)
        code_generator.swap_reverse_stack_items(2)

        code_generator.convert_end_if(is_equal)
        for address in jmps_to_condition:
            code_generator.convert_end_if(address, is_internal=True)

        #   index -= 1
        code_generator.insert_opcode(Opcode.DEC)

        # while condition
        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.SIGN)
        code_generator.convert_literal(0)
        code_generator.convert_operation(BinaryOp.GtE)

        code_generator.convert_end_while(while_start, while_condition, is_internal=True)

        # clean stack
        code_generator.remove_stack_top_item()
        code_generator.swap_reverse_stack_items(3)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()

    def _generic_verification(self, code_generator) -> tuple[list[int], list[int]]:
        """
        Generate the Neo VM opcodes for the method.

        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        return [], []



from typing import Any

from boa3.internal.model.type.enum.intenumtype import IntEnumType
from boa3.internal.model.type.itype import IType
from boa3.internal.neo3.contracts.native import Role


class RoleType(IntEnumType):
    """
    A class used to represent Neo's Role type.
    """

    def __init__(self):
        super().__init__(Role)
        self._identifier = 'Role'

    @classmethod
    def build(cls, value: Any = None) -> IType:
        if value is None or cls._is_type_of(value):
            return _Role

    @classmethod
    def _is_type_of(cls, value: Any):
        return isinstance(value, (Role, RoleType))


_Role = RoleType()



__all__ = [
    'Nep17InterfaceBalanceOfMethod',
    'Nep17InterfaceDecimalsMethod',
    'Nep17InterfaceSymbolMethod',
    'Nep17InterfaceTotalSupplyMethod',
    'Nep17InterfaceTransferMethod',
]

from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacebalanceofmethod import Nep17InterfaceBalanceOfMethod
from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacedecimalsmethod import Nep17InterfaceDecimalsMethod
from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacesymbolmethod import Nep17InterfaceSymbolMethod
from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacetotalsupplymethod import Nep17InterfaceTotalSupplyMethod
from boa3.internal.model.builtin.contract.nep17_interface_methods.nep17interfacetransfermethod import Nep17InterfaceTransferMethod



import ast
from abc import ABC

from boa3.internal.model.builtin.builtincallable import IBuiltinCallable
from boa3.internal.model.event import Event
from boa3.internal.model.variable import Variable


class IBuiltinEvent(IBuiltinCallable, Event, ABC):
    def __init__(self,
                 identifier: str,
                 args: dict[str, Variable] = None,
                 defaults: list[ast.AST] = None,
                 vararg: tuple[str, Variable] | None = None,
                 kwargs: dict[str, Variable] | None = None,
                 deprecated: bool = False
                 ):
        from boa3.internal.model.type.type import Type
        super().__init__(identifier, args, vararg, kwargs, defaults, Type.none, deprecated)

        # constructor of IBuiltinCallable and Event classes are conflicting
        self._identifier = identifier
        self.name = identifier
        self.args = args if args is not None else {}
        self.defaults = defaults if defaults is not None else []



__all__ = [
    'PolicyContract',
]

from boa3.internal.neo3.network.payloads.transactionattributetype import TransactionAttributeType
from boa3.sc.types import UInt160


class PolicyContract:
    """
    A class used to represent the Policy native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/Policy>`__
    to learn more about the Policy class.
    """

    hash: UInt160

    @classmethod
    def get_fee_per_byte(cls) -> int:
        """
        Gets the network fee per transaction byte.

        >>> PolicyContract.get_fee_per_byte()
        1000

        :return: the network fee per transaction byte
        :rtype: int
        """
        pass

    @classmethod
    def get_exec_fee_factor(cls) -> int:
        """
        Gets the execution fee factor. This is a multiplier that can be adjusted by the committee to adjust the system fees
        for transactions.

        >>> PolicyContract.get_exec_fee_factor()
        30

        :return: the execution fee factor
        :rtype: int
        """
        pass

    @classmethod
    def get_storage_price(cls) -> int:
        """
        Gets the storage price.

        >>> PolicyContract.get_storage_price()
        100000

        :return: the storage price
        :rtype: int
        """
        pass

    @classmethod
    def get_attribute_fee(cls, attribute_type: TransactionAttributeType) -> int:
        """
        Gets the fee for attribute.

        >>> PolicyContract.get_attribute_fee(TransactionAttributeType.HIGH_PRIORITY)
        0

        :return: the fee for attribute
        :rtype: int
        """
        pass

    @classmethod
    def is_blocked(cls, account: UInt160) -> bool:
        """
        Determines whether the specified account is blocked.

        >>> PolicyContract.is_blocked(UInt160(b'\\xcfv\\xe2\\x8b\\xd0\\x06,JG\\x8e\\xe3Ua\\x01\\x13\\x19\\xf3\\xcf\\xa4\\xd2'))
        False

        :param account: the account to be checked
        :type account: boa3.sc.types.UInt160

        :return: whether the account is blocked or not
        :rtype: bool
        """
        pass

    @classmethod
    def set_attribute_fee(cls, attribute_type: TransactionAttributeType, value: int) -> None:
        """
        Sets the fee for attribute. You need to sign the transaction using a committee member, otherwise, this function
        will throw an error.

        >>> PolicyContract.set_attribute_fee(TransactionAttributeType.HIGH_PRIORITY, 10)
        None
        """
        pass



__all__ = ['GetContractByIdMethod',
           'HasMethod',
           ]

from boa3.internal.model.builtin.native.contract_management.getcontractbyid import GetContractByIdMethod
from boa3.internal.model.builtin.native.contract_management.hasmethod import HasMethod



from boa3.internal.model.operation.binary.binaryoperation import BinaryOperation
from boa3.internal.model.operation.operator import Operator
from boa3.internal.model.type.type import IType, Type


class ElvisOperatorOr(BinaryOperation):
    """
    A class used to represent the boolean or operation

    :ivar operator: the operator of the operation. Inherited from :class:`IOperation`
    :ivar left: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar right: the left operand type. Inherited from :class:`BinaryOperation`
    :ivar result: the result type of the operation.  Inherited from :class:`IOperation`
    """
    _valid_types: list[IType] = [Type.bool]

    def __init__(self, left: IType = Type.bool, right: IType = None):
        self.operator: Operator = Operator.Or
        super().__init__(left, right)

    def validate_type(self, *types: IType) -> bool:
        if len(types) != self.number_of_operands:
            return False
        left: IType = types[0]
        right: IType = types[1]

        return True

    def _get_result(self, left: IType, right: IType) -> IType:
        if left.is_type_of(right) and right.is_type_of(left):
            return right
        else:
            return Type.union.build([left, right])

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin

        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Bool, is_internal=True)

        is_true = code_generator.convert_begin_if()
        code_generator.remove_stack_item(2)

        is_false = code_generator.convert_begin_else(is_true, is_internal=True)
        code_generator.remove_stack_top_item()
        code_generator.convert_end_if(is_false, is_internal=True)



import ast


class VariableGenerationData:
    def __init__(self, var_id: str, index: ast.AST | None, address: int):
        self.var_id: str = var_id
        self.index: ast.AST | None = index
        self.address: int = address

    def __iter__(self):
        yield self.var_id
        yield self.index
        yield self.address



from abc import abstractmethod
from typing import Any, Self

from boa3.internal.model.identifiedsymbol import IdentifiedSymbol
from boa3.internal.model.symbol import ISymbol
from boa3.internal.neo.vm.type.AbiType import AbiType
from boa3.internal.neo.vm.type.StackItem import StackItemType


class IType(IdentifiedSymbol):
    """
    An interface used to represent types

    :ivar identifier: the name identifier of the type
    """

    def __init__(self, identifier: str, deprecated: bool = False):
        super().__init__(identifier, deprecated)

    @property
    def shadowing_name(self) -> str:
        return 'type'

    def __repr__(self) -> str:
        return self.__str__()

    def __str__(self) -> str:
        return self.identifier

    @property
    def default_value(self) -> Any:
        return None

    @property
    def abi_type(self) -> AbiType:
        """
        Get the type representation for the abi

        :return: the representation for the abi. Any by default.
        """
        return AbiType.Any

    @property
    def stack_item(self) -> StackItemType:
        """
        Get the Neo VM stack item type representation for this type

        :return: the stack item type of this type. Any by default.
        """
        return StackItemType.Any

    @property
    def is_generic(self) -> bool:
        """
        Verifies if this is a generic type

        :return: True if this is a generic type. False otherwise.
        """
        return False

    @classmethod
    @abstractmethod
    def _is_type_of(cls, value: Any) -> bool:
        """
        Validates if this is the type of the given value

        :param value: value to check the type
        :return: a boolean value that represents if this is the type of the value.
        :rtype: bool
        """
        pass

    def is_type_of(self, value: Any) -> bool:
        """
        Validates if this is the type of the given value. If this is not a primitive type, validates if its attributes
        are also valid.

        :param value: value to check the type
        :return: a boolean value that represents if this is the type of the value.
        :rtype: bool
        """
        return self._is_type_of(value)

    @classmethod
    @abstractmethod
    def build(cls, value: Any) -> Self:
        """
        Creates a type instance with the given value

        :param value: value to build the type
        :return: The built type if the value is valid. None otherwise
        :rtype: IType or None
        """
        pass

    def generate_is_instance_type_check(self, code_generator):
        """
        Generates the opcodes to check if a value is of this type

        :type code_generator: boa3.internal.compiler.codegenerator.codegenerator.CodeGenerator
        """
        code_generator.insert_type_check(self.stack_item)

    @property
    def symbols(self) -> dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the class type with its name
        """
        return {}

    def union_type(self, other_type: "IType") -> "IType":
        """
        Gets a type that is an union of `self` and `other_type`

        :param other_type: type that'll be united with `self`
        :type other_type: IType
        :return: an union of this type and other_type
        :rtype: IType
        """
        from boa3.internal.model.type.annotation.uniontype import UnionType
        return UnionType.build((self, other_type))

    def except_type(self, other_type: "IType") -> "IType":
        """
        Gets a type that is type of `self` but is not type of `other_type`.
        If `other_type` is an implementation of `self`, returns `self`

        :param other_type: type that won't be allowed in the resulting type
        :type other_type: IType
        :return: an union of this type and the exclusion of other_type
        :rtype: IType
        """
        return self

    def intersect_type(self, other_type: "IType") -> "IType":
        """
        Gets a type that is the intersection of `self` but is not type of `other_type`.
        If `other_type` is an implementation of `self`, returns `other_type`
        If `other_type` and `self` has no values in common, returns None

        :param other_type: type that'll be intersected with self
        :type other_type: IType
        :return: an intersection of this type and the exclusion of other_type
        :rtype: IType
        """
        if self.is_type_of(other_type):
            return other_type
        if other_type.is_type_of(self):
            return self

        from boa3.internal.model.type.annotation.uniontype import UnionType
        if isinstance(other_type, UnionType):
            same_types = [x for x in other_type.union_types if self.is_type_of(x)]
            if len(same_types) == 0:
                return self
            elif len(same_types) == 1:
                return same_types[0]
            else:
                return UnionType.build(same_types)

        from boa3.internal.model.type.type import Type
        return Type.none

    def is_equal(self, other: Any) -> bool:
        if not isinstance(other, IType):
            return False
        return self.is_type_of(other) and other.is_type_of(self)



from enum import IntFlag
from functools import reduce
from typing import Type as TType, Any

from boa3.internal.model.builtin.method import IBuiltinMethod
from boa3.internal.model.expression import IExpression
from boa3.internal.model.method import Method
from boa3.internal.model.symbol import ISymbol
from boa3.internal.model.type.primitive.inttype import IntType
from boa3.internal.model.type.type import Type
from boa3.internal.model.variable import Variable


class IntFlagType(IntType):
    """
    A class used to represent Python IntFlag type
    """

    def __init__(self, enum_type: TType[IntFlag]):
        super().__init__()
        self._identifier = '-IntFlag'
        self._enum_type = enum_type

    @property
    def symbols(self) -> dict[str, ISymbol]:
        """
        Gets the class symbols of this type

        :return: a dictionary that maps each symbol in the module with its name
        """
        _symbols = super().symbols
        _symbols.update({name: Variable(self) for name in self._enum_type.__members__.keys()})

        return _symbols

    def get_value(self, symbol_id) -> Any:
        """
        Gets the literal value of a symbol

        :return: the value if this type has this symbol. None otherwise.
        """
        if symbol_id in self.symbols and symbol_id in self._enum_type.__members__:
            return self._enum_type.__members__[symbol_id]

        return None

    def constructor_method(self) -> Method | None:
        if self._constructor is None:
            self._constructor: Method = IntEnumMethod(self, self._enum_type)
        return self._constructor

    @property
    def exception_message(self) -> str:
        return f"Invalid {self.identifier} parameter value"

    @property
    def all_flags(self) -> IntFlag:
        flag_values = [value for value in self._enum_type]
        return reduce(lambda x, y: x | y, flag_values)

    @property
    def next_power_of_two(self) -> int:
        """
        Gets the next power of two value
        """
        highest_enum = self.all_flags
        next_power_of_two = 1
        while next_power_of_two < highest_enum:
            next_power_of_two *= 2

        return next_power_of_two

class IntEnumMethod(IBuiltinMethod):

    def __init__(self, return_type: IntFlagType, enum_type: TType[IntFlag]):
        identifier = f'-{return_type.identifier}__init__'
        args: dict[str, Variable] = {
            'x': Variable(Type.int)
        }
        super().__init__(identifier, args, return_type=return_type)
        self._enum_type = enum_type

    def validate_parameters(self, *params: IExpression) -> bool:
        return len(params) == 1

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return None

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.operation.unaryop import UnaryOp

        code_generator.duplicate_stack_top_item()
        code_generator.convert_literal(self.return_type.all_flags)
        code_generator.convert_operation(UnaryOp.BitNot, is_internal=True)
        code_generator.convert_operation(BinaryOp.BitAnd, is_internal=True)

        code_generator.convert_operation(UnaryOp.Not, is_internal=True)
        code_generator.convert_literal(self.return_type.exception_message)
        code_generator.convert_assert(has_message=True)



from boa3.internal.model.builtin.interop.storage.put.storageputboolmethod import StoragePutBoolMethod
from boa3.internal.model.builtin.interop.storage.put.storageputbytesmethod import StoragePutBytesMethod
from boa3.internal.model.builtin.interop.storage.put.storageputdictmethod import StoragePutDictMethod
from boa3.internal.model.builtin.interop.storage.put.storageputecpointmethod import StoragePutECPointMethod
from boa3.internal.model.builtin.interop.storage.put.storageputintmethod import StoragePutIntMethod
from boa3.internal.model.builtin.interop.storage.put.storageputlistmethod import StoragePutListMethod
from boa3.internal.model.builtin.interop.storage.put.storageputobjectmethod import StoragePutObjectMethod
from boa3.internal.model.builtin.interop.storage.put.storageputstrmethod import StoragePutStrMethod
from boa3.internal.model.builtin.interop.storage.put.storageputuint160method import StoragePutUInt160Method
from boa3.internal.model.builtin.interop.storage.put.storageputuint256method import StoragePutUInt256Method



from boa3.internal.model.builtin.classmethod.popmethod import PopMethod
from boa3.internal.model.type.itype import IType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class PopDictDefaultMethod(PopMethod):

    def __init__(self, arg_value: IType | None = None):
        from boa3.internal.model.type.type import Type

        if not Type.dict.is_type_of(arg_value):
            arg_value = Type.dict

        args: dict[str, Variable] = {
            'self': Variable(arg_value),
            'key': Variable(arg_value.valid_key),
            'default': Variable(Type.any)
        }

        super().__init__(args, return_type=arg_value.value_type)

    def generate_opcodes(self, code_generator):
        from boa3.internal.model.type.type import Type

        # put the `default` value at the bottom of the stack
        code_generator.swap_reverse_stack_items(3)
        code_generator.swap_reverse_stack_items(2)
        code_generator.swap_reverse_stack_items(3)

        code_generator.duplicate_stack_item(2)
        code_generator.duplicate_stack_item(2)
        code_generator.insert_opcode(Opcode.HASKEY, pop_from_stack=True, add_to_stack=[Type.bool])
        # if the key is in the dictionary:
        if_has_key = code_generator.convert_begin_if()
        #   return the dict[key] and remove pair from dict
        self.generate_internal_opcodes(code_generator)
        code_generator.remove_stack_item(2)

        # else the key is not in the dictionary:
        else_not_has_key = code_generator.convert_begin_else(if_has_key, is_internal=True)
        code_generator.remove_stack_top_item()
        code_generator.remove_stack_top_item()
        # return the default value
        code_generator.convert_end_if(else_not_has_key, is_internal=True)



from boa3.internal.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.internal.model.variable import Variable


class RecoverSecp256K1Method(CryptoLibMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'recover_secp256k1'
        syscall = 'recoverSecp256K1'
        args: dict[str, Variable] = {
            'message_hash': Variable(Type.bytes),
            'signature': Variable(Type.bytes),
        }

        super().__init__(identifier, syscall, args, return_type=Type.optional.build(Type.bytes))



from boa3.internal.model.builtin.interop.storage.put.istorageputmethod import IStoragePutMethod


class StoragePutIntMethod(IStoragePutMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type

        identifier = 'put_int'
        value_type = Type.int

        super().__init__(identifier, value_type=value_type)

    def generate_serialize_value_opcodes(self, code_generator):
        # it doesn't need to serialize primitive types values on put
        return



from collections.abc import Iterable
from typing import Any

from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.type.collection.sequence.sequencetype import SequenceType
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class ReversedMethod(IBuiltinMethod):
    def __init__(self, args_type: SequenceType = None):
        from boa3.internal.model.type.type import Type
        identifier = 'reversed'
        if not isinstance(args_type, SequenceType):
            args_type = Type.sequence

        args: dict[str, Variable] = {'sequence': Variable(args_type)}

        super().__init__(identifier, args, return_type=Type.reversed)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def identifier(self) -> str:
        from boa3.internal.model.type.type import Type
        if self.args['sequence'].type is Type.sequence:
            return self.raw_identifier
        return '-{0}_from_{1}'.format(self._identifier, self.args['sequence'].type.identifier)

    def build(self, value: Any) -> IBuiltinMethod:
        if isinstance(value, Iterable) and len(value) > 0:
            value = value[0]

        if not isinstance(value, SequenceType):
            return ReversedMethod()

        if type(value) == type(self.args['sequence'].type):
            return self

        return ReversedMethod(value)

    @property
    def _body(self) -> str | None:
        return

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.model.operation.binaryop import BinaryOp
        from boa3.internal.model.type.type import Type
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        # if isinstance(arg1, (str, bytes)):
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.ByteString)
        is_byte_str = code_generator.convert_begin_if()

        #     list_aux = []
        code_generator.convert_new_empty_array(length=0, array_type=self.return_type)
        #     limit = len(arg)
        code_generator.swap_reverse_stack_items(2)
        code_generator.duplicate_stack_top_item()
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)
        #     index = 0
        code_generator.convert_literal(0)

        #     while index < limit:
        while_index_is_valid = code_generator.convert_begin_while()
        #         list_aux.append(arg[index])
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(4)
        code_generator.duplicate_stack_item(3)
        code_generator.convert_literal(1)
        code_generator.convert_get_substring(is_internal=True)
        code_generator.convert_cast(Type.int if Type.bytes.is_type_of(self.args['sequence'].type)
                                    else Type.str, is_internal=True)
        code_generator.convert_builtin_method_call(Builtin.SequenceAppend, is_internal=True)
        #         index += 1
        code_generator.insert_opcode(Opcode.INC)

        while_condition = code_generator.bytecode_size
        code_generator.duplicate_stack_top_item()
        code_generator.duplicate_stack_item(3)
        code_generator.convert_operation(BinaryOp.Lt, is_internal=True)
        code_generator.convert_end_while(while_index_is_valid, while_condition, is_internal=True)

        for _ in range(3):
            code_generator.remove_stack_top_item()

        # else:
        else_is_bytes_str = code_generator.convert_begin_else(is_byte_str, is_internal=True)
        #     list_aux = arg.copy()
        code_generator.convert_copy()

        # list_aux.reverse()
        code_generator.convert_end_if(else_is_bytes_str, is_internal=True)

        code_generator.duplicate_stack_top_item()
        code_generator.insert_opcode(Opcode.REVERSEITEMS)



from boa3.internal.model.builtin.interop.nativecontract import CryptoLibMethod
from boa3.internal.model.variable import Variable


class Bls12381AddMethod(CryptoLibMethod):

    def __init__(self):
        from boa3.internal.model.builtin.interop.crypto import Bls12381Type
        bls12_381_type = Bls12381Type.build()

        identifier = 'bls12_381_add'
        native_identifier = 'bls12381Add'
        args: dict[str, Variable] = {
            'x': Variable(bls12_381_type),
            'y': Variable(bls12_381_type),
        }
        super().__init__(identifier, native_identifier, args, return_type=bls12_381_type)



__all__ = [
    'GasToken',
]


from typing import Any

from boa3.sc.types import UInt160


class GasToken:
    """
    A class used to represent the GAS native contract.

    Check out `Neo's Documentation <https://developers.neo.org/docs/n3/reference/scapi/framework/native/Gas>`__
    to learn more about the GAS class.
    """

    hash: UInt160

    @classmethod
    def symbol(cls) -> str:
        """
        Gets the symbol of GAS.

        >>> GasToken.symbol()
        'GAS'

        :return: the GAS string.
        :rtype: str
        """
        pass

    @classmethod
    def decimals(cls) -> int:
        """
        Gets the amount of decimals used by GAS.

        >>> GasToken.decimals()
        8

        :return: the number 8.
        :rtype: int
        """
        pass

    @classmethod
    def totalSupply(cls) -> int:
        """
        Gets the total token supply deployed in the system.

        >>> GasToken.totalSupply()
        5199999098939320

        >>> GasToken.totalSupply()
        5522957322800300

        :return: the total token supply deployed in the system.
        :rtype: int
        """
        pass

    @classmethod
    def balanceOf(cls, account: UInt160) -> int:
        """
        Get the current balance of an address.

        >>> GasToken.balanceOf(UInt160(bytes(20)))
        0

        >>> GasToken.balanceOf(UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'))
        1000000000

        :param account: the account's address to retrieve the balance for
        :type account: boa3.sc.types.UInt160
        :return: the account's balance
        :rtype: int
        """
        pass

    @classmethod
    def transfer(cls, from_address: UInt160, to_address: UInt160, amount: int, data: Any = None) -> bool:
        """
        Transfers an amount of GAS from one account to another.

        If the method succeeds, it will fire the `Transfer` event and must return true, even if the amount is 0,
        or from and to are the same address.

        >>> GasToken.transfer(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'),     # this script hash needs to have signed the transaction or block
        ...              UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'),
        ...              10000, None)
        True

        >>> GasToken.transfer(UInt160(bytes(20)),
        ...              UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'),
        ...              10000, None)
        False

        >>> GasToken.transfer(UInt160(b'\\xc9F\\x17\\xba!\\x99\\x07\\xc1\\xc5\\xd6\t#\\xe1\\x9096\\x89U\\xac\\x13'),
        ...              UInt160(b'\\xabv\\xe2\\xcb\\xb0\\x16,vG\\x2f\\x44Va\\x10\\x14\\x19\\xf3\\xff\\xa1\\xe6'),
        ...              -1, None)
        False

        :param from_address: the address to transfer from
        :type from_address: boa3.sc.types.UInt160
        :param to_address: the address to transfer to
        :type to_address: boa3.sc.types.UInt160
        :param amount: the amount of GAS to transfer
        :type amount: int
        :param data: whatever data is pertinent to the onNEP17Payment method
        :type data: Any

        :return: whether the transfer was successful
        :rtype: bool
        :raise Exception: raised if `from_address` or `to_address` length is not 20 or if `amount` is less than zero.
        """
        pass



from boa3.internal import constants
from boa3.internal.neo import to_hex_str
from boa3_test.test_drive.model.smart_contract.testcontract import TestContract
from boa3_test.test_drive.testrunner.blockchain.contract import TestRunnerContract


class ContractCollection:

    _native_contracts = [
        TestRunnerContract('OracleContract', to_hex_str(constants.ORACLE_SCRIPT)),
        TestRunnerContract('RoleManagement', to_hex_str(constants.ROLE_MANAGEMENT)),
        TestRunnerContract('PolicyContract', to_hex_str(constants.POLICY_SCRIPT)),
        TestRunnerContract('GasToken', to_hex_str(constants.GAS_SCRIPT)),
        TestRunnerContract('NeoToken', to_hex_str(constants.NEO_SCRIPT)),
        TestRunnerContract('LedgerContract', to_hex_str(constants.LEDGER_SCRIPT)),
        TestRunnerContract('CryptoLib', to_hex_str(constants.CRYPTO_SCRIPT)),
        TestRunnerContract('StdLib', to_hex_str(constants.STD_LIB_SCRIPT)),
        TestRunnerContract('ContractManagement', to_hex_str(constants.MANAGEMENT_SCRIPT)),
    ]

    def __init__(self):
        natives = self._native_contracts.copy()

        self._contract_names: list[str] = [native.name for native in natives]
        self._contract_paths: list[str] = [native.path for native in natives]
        self._internal_list: list[TestContract] = natives
        self._waiting_deploy: list[TestContract] = []

    def append(self, new_contract: TestContract):
        if not isinstance(new_contract, TestContract):
            return
        if new_contract.name not in self._contract_names:
            self._contract_names.append(new_contract.name)
            self._contract_paths.append(new_contract.path)
            if new_contract.script_hash is None:
                self._waiting_deploy.append(new_contract)
            return self._internal_list.append(new_contract)

    def remove(self, contract: TestContract):
        try:
            contract_index = self._contract_names.index(contract.name)
            self._contract_names.pop(contract_index)
            self._contract_paths.pop(contract_index)
            contract = self._internal_list.pop(contract_index)
            if contract.script_hash is None:
                self._waiting_deploy.remove(contract)
        except ValueError:
            return

    def pop(self, index: int = -1):
        self._contract_names.pop(index)
        self._contract_paths.pop(index)
        contract = self._internal_list.pop(index)
        if contract.script_hash is None:
            self._waiting_deploy.remove(contract)
        return contract

    def clear(self):
        self._contract_names.clear()
        self._contract_paths.clear()
        self._waiting_deploy.clear()
        return self._internal_list.clear()

    def __len__(self):
        return len(self._internal_list)

    def __contains__(self, item) -> bool:
        if isinstance(item, str):
            if item in self._contract_names:
                return True

            if item in self._contract_paths:
                return True

        if hasattr(item, 'name') and item.name in self._contract_names:
            return True

        return any(contract.is_valid_identifier(item) for contract in self._internal_list)

    def __getitem__(self, item) -> TestContract:
        if isinstance(item, str):
            try:
                try:
                    contract_index = self._contract_names.index(item)
                except ValueError:
                    contract_index = self._contract_paths.index(item)

                return self._internal_list[contract_index]
            except ValueError:
                pass

        return next((contract for contract in self._internal_list if contract.is_valid_identifier(item)), None)

    def __str__(self):
        return str(self._internal_list)

    def __repr__(self):
        return self._internal_list.__repr__()

    def replace(self, deployed_contracts: list[TestContract]):
        already_existing_contracts = self._contract_names.copy()
        contract_indexes = list(range(len(already_existing_contracts)))

        for contract in deployed_contracts:
            try:
                contract_index = self._contract_names.index(contract.name)
                existing_contract = self._internal_list[contract_index]
                if existing_contract.script_hash is None:
                    existing_contract.script_hash = contract.script_hash
                    self._waiting_deploy.remove(existing_contract)

                already_existing_contracts.remove(contract.name)
                contract_indexes.remove(contract_index)
            except ValueError:
                self.append(contract)

        # these contracts were destroyed in the test blockchain
        for removed_index in reversed(contract_indexes):
            self._contract_names.pop(removed_index)
            self._contract_paths.pop(removed_index)
            contract = self._internal_list.pop(removed_index)
            if contract.script_hash is None:
                self._waiting_deploy.remove(contract)

    def update_after_deploy(self):
        for contract in self._waiting_deploy.copy():
            script_hash = contract.script_hash
            if script_hash is not None:
                self._waiting_deploy.remove(contract)



from boa3.internal.compiler.compiler import Compiler
from boa3.internal.exception.InvalidPathException import InvalidPathException

__all__ = ['Boa3']


class Boa3:
    """
    The main class.
    Contains the methods that the final user have access to.
    """

    @staticmethod
    def compile(
            path: str,
            root_folder: str = None,
            log_level: str = None,
            env: str = None,
            fail_fast: bool = True,
            optimize: bool = True
    ) -> bytes:
        """
        Load a Python file to be compiled but don't write the result into a file

        :param path: the path of the Python file to compile
        :param root_folder: the root path of the project
        :param env: specific environment id to compile
        :param fail_fast: if should stop compilation on first error found.
        :return: the bytecode of the compiled .nef file
        """
        if not path.endswith('.py'):
            raise InvalidPathException(path)

        return Compiler().compile(path, root_folder, env,
                                  log_level=log_level,
                                  fail_fast=fail_fast,
                                  optimize=optimize
                                  )

    @staticmethod
    def compile_and_save(
            path: str,
            output_path: str = None,
            root_folder: str = None,
            show_errors: bool = True,
            log_level: str = None,
            debug: bool = False,
            env: str = None,
            fail_fast: bool = True,
            optimize: bool = True
    ):
        """
        Load a Python file to be compiled and save the result into the files.
        By default, the resultant .nef file is saved in the same folder of the
        source file.

        :param path: the path of the Python file to compile
        :param output_path: Optional path to save the generated files
        :param root_folder: the root path of the project
        :param show_errors: if compiler errors should be logged.
        :param debug: if nefdbgnfo file should be generated.
        :param env: specific environment id to compile.
        :param fail_fast: if should stop compilation on first error found.
        """
        if not path.endswith('.py'):
            raise InvalidPathException(path)

        if output_path is None:
            output_path = path.replace('.py', '.nef')
        elif not output_path.endswith('.nef'):
            raise InvalidPathException(output_path)

        Compiler().compile_and_save(path, output_path, root_folder, show_errors, log_level, debug, env, fail_fast,
                                    optimize
                                    )



__all__ = [
    'ExecutionScript'
]

from typing import Self

from boa3.internal.compiler.codegenerator.vmcodemapping import VMCodeMapping
from boa3.internal.neo.vm.VMCode import VMCode
from boa3.internal.neo3.contracts.nef import MethodToken


class ExecutionScript:
    def __init__(self, code_map: dict[int, VMCode], tokens: list[MethodToken]):
        self._addresses = tuple(code_map.keys())
        self._instructions = tuple(code_map.values())

        self._last_address = self._addresses[-1] if len(self._addresses) else -1
        self._last_instruction_size = self._instructions[-1].size if len(self._instructions) else 0

        self._tokens = tokens

    @classmethod
    def from_code_map(cls, instance: VMCodeMapping) -> Self:
        obj = ExecutionScript(instance.code_map, instance._method_tokens.to_list())
        return obj

    def get_instruction(self, address) -> VMCode:
        if address not in self._addresses:
            raise IndexError
        index = self._addresses.index(address)
        return self._instructions[index]

    def next_address(self, address) -> int:
        if address in self._addresses:
            address_index = self._addresses.index(address)
            if address < self._last_address:
                next_address = self._addresses[address_index + 1]
            else:
                next_address = self._last_address + self._last_instruction_size
        else:
            next_address = next((instr_address
                                 for instr_address in self._addresses
                                 if instr_address > address),
                                self._last_address + self._last_instruction_size)

        return next_address

    def __contains__(self, obj) -> bool:
        return obj in self._addresses



from boa3.internal.model.builtin.interop.blockchain import SignerType
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.variable import Variable

class GetCurrentSignersMethod(InteropMethod):
    def __init__(self, signer_type: SignerType):
        from boa3.internal.model.type.type import Type
        identifier = 'get_current_signers'
        syscall = 'System.Runtime.CurrentSigners'
        args: dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=Type.list.build([signer_type]))



from boa3.internal.model.builtin.method.builtinmethod import IBuiltinMethod
from boa3.internal.model.variable import Variable
from boa3.internal.neo.vm.opcode.Opcode import Opcode


class BoolMethod(IBuiltinMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'bool'

        args: dict[str, Variable] = {
            'value': Variable(Type.any),
        }
        super().__init__(identifier, args, return_type=Type.bool)

    @property
    def _value(self) -> Variable:
        return self.args['value']

    def generate_internal_opcodes(self, code_generator):
        from boa3.internal.model.builtin.builtin import Builtin
        from boa3.internal.neo.vm.type.StackItem import StackItemType

        # if arg is None
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(None)
        is_null = code_generator.convert_begin_if()
        #   return False
        code_generator.remove_stack_top_item()
        code_generator.convert_literal(False)

        else_is_null = code_generator.convert_begin_else(is_null, is_internal=True)
        # if arg is not int
        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.Array)
        if_is_sequence = code_generator.convert_begin_if()
        code_generator.change_jump(if_is_sequence, Opcode.JMPIF)

        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.Map)
        if_is_map = code_generator.convert_begin_if()
        code_generator.change_jump(if_is_map, Opcode.JMPIF)

        code_generator.duplicate_stack_top_item()
        code_generator.insert_type_check(StackItemType.Struct)
        if_is_struct = code_generator.convert_begin_if()

        code_generator.convert_end_if(if_is_sequence, is_internal=True)
        code_generator.convert_end_if(if_is_map, is_internal=True)
        #   return len(arg) == 0
        code_generator.convert_builtin_method_call(Builtin.Len, is_internal=True)

        code_generator.convert_end_if(if_is_struct, is_internal=True)
        # else
        #   return arg == 0
        code_generator.convert_end_if(else_is_null, is_internal=True)

        code_generator.insert_opcode(Opcode.NZ)

    @property
    def _args_on_stack(self) -> int:
        return len(self.args)

    @property
    def _body(self) -> str | None:
        return



import re

from boa3.internal.analyser import supportedstandard
from boa3.internal.analyser.astanalyser import IAstAnalyser
from boa3.internal.exception import CompilerError
from boa3.internal.model.event import Event
from boa3.internal.model.method import Method
from boa3.internal.model.symbol import ISymbol


class StandardAnalyser(IAstAnalyser):
    """
    This class is responsible for the checking if the given contract standards are implemented

    :ivar symbols: a dictionary that maps the global symbols.
    """

    def __init__(self, analyser, symbol_table: dict[str, ISymbol],
                 log: bool = False, fail_fast: bool = True):
        from boa3.sc.compiletime import NeoMetadata

        super().__init__(analyser.ast_tree, analyser.filename, analyser.root, log=log, fail_fast=fail_fast)

        self.symbols: dict[str, ISymbol] = symbol_table

        if isinstance(analyser.metadata, NeoMetadata):
            # filter only strings
            analyser.metadata.supported_standards = [standard for standard in analyser.metadata.supported_standards
                                                     if isinstance(standard, str)]
            standards = analyser.metadata.supported_standards
        else:
            standards = []

        self.standards: list[str] = standards
        self._analyser = analyser
        self._filter_standards_names()
        self._validate_standards()
        self._check_other_implemented_standards()

    def _filter_standards_names(self):
        """
        Converts all the Neo standards names to upper kebab case
        """
        filtered_standards = set()
        for standard in self.standards:
            standard = standard.strip()
            if standard.upper().startswith('NEP'):
                rebab_case = re.sub(r'([A-Z]+|(.\d+)+|\d+)[^a-zA-Z\d]?', r'\g<1>-', standard.upper())
                if rebab_case.endswith('-'):
                    rebab_case = rebab_case[:-1]
                standard = rebab_case

            filtered_standards.add(standard)

        self.standards.clear()
        self.standards.extend(filtered_standards)

    def get_methods_by_display_name(self, method_id: str) -> list[Method]:
        methods = []
        for symbol_id, symbol in self.symbols.items():
            if isinstance(symbol, Method) and symbol.is_public:
                display_name = symbol.external_name if symbol.external_name is not None else symbol_id
                if display_name == method_id:
                    methods.append(symbol)
        return methods

    def _validate_standards(self):
        try:
            for standard in self.standards:
                if standard in supportedstandard.neo_standards:
                    standard_was_found = False
                    standard_errors = {}

                    standard_index = 0
                    while standard_index < len(supportedstandard.neo_standards[standard]) and not standard_was_found:
                        current_standard = supportedstandard.neo_standards[standard][standard_index]
                        current_standard_errors = []

                        # validate standard's methods
                        for standard_method in current_standard.methods:
                            method_id = standard_method.external_name
                            is_implemented = False

                            found_methods = self.get_methods_by_display_name(method_id)
                            for method in found_methods:
                                if isinstance(method, Method) and current_standard.match_definition(standard_method, method):
                                    is_implemented = True
                                    break

                            if not is_implemented:
                                current_standard_errors.append(
                                    CompilerError.MissingStandardDefinition(standard, method_id, standard_method)
                                )

                        # validate standard's events
                        events = [symbol for symbol in self.symbols.values() if isinstance(symbol, Event)]
                        # imported events should be included in the validation
                        for import_ in self._analyser.get_imports():
                            events.extend([event for event in import_.symbol_table.values()
                                           if isinstance(event, Event) and event not in events])

                        for standard_event in current_standard.events:
                            is_implemented = False
                            for event in events:
                                if (event.name == standard_event.name
                                        and current_standard.match_definition(standard_event, event)):
                                    is_implemented = True
                                    break

                            if not is_implemented:
                                current_standard_errors.append(
                                    CompilerError.MissingStandardDefinition(standard,
                                                                            standard_event.name,
                                                                            standard_event)
                                )

                        # validate optional methods
                        for optional_method in current_standard.optionals:
                            method_id = optional_method.external_name
                            is_implemented = False

                            found_methods = self.get_methods_by_display_name(method_id)
                            for method in found_methods:
                                if isinstance(method, Method) and current_standard.match_definition(optional_method, method):
                                    is_implemented = True
                                    break

                            if found_methods and not is_implemented:
                                current_standard_errors.append(
                                    CompilerError.MissingStandardDefinition(standard, method_id, optional_method)
                                )

                        standard_errors[standard_index] = current_standard_errors
                        standard_was_found = len(current_standard_errors) == 0
                        standard_index += 1

                    if not standard_was_found:
                        for error in standard_errors[0]:
                            self._log_error(error)

        except CompilerError.CompilerError:
            # stops the analyser if fail fast is activated
            pass

    def _check_other_implemented_standards(self):
        other_standards = supportedstandard.neo_standards.copy()
        # verify only standards that were not mentioned
        for standard in self.standards:
            if standard in supportedstandard.neo_standards:
                other_standards.pop(standard)

        # gets a list of all events
        events = [symbol for symbol in self.symbols.copy().values() if isinstance(symbol, Event)]
        for import_ in self._analyser.get_imports():
            events.extend([event for event in import_.symbol_table.values() if isinstance(event, Event)])

        # verify if the methods and events that were implemented corresponds to a standard
        for standard in other_standards:
            for current_standard in other_standards[standard]:
                # verify the methods
                methods_implemented = True
                standard_methods = current_standard.methods
                index = 0

                while methods_implemented and index < len(standard_methods):
                    standard_method = standard_methods[index]
                    method_id = standard_method.external_name
                    found_methods = self.get_methods_by_display_name(method_id)

                    methods_implemented = any(
                        current_standard.match_definition(standard_method, method) for method in found_methods
                    )
                    index += 1

                if not methods_implemented:
                    continue    # if even one of the methods was not implemented, then check the next standard

                # verify the events
                events_implemented = True
                standard_events = current_standard.events
                index = 0

                while events_implemented and index < len(standard_events):
                    standard_event = standard_events[index]
                    events_implemented = any(
                        (event.name == standard_event.name and
                         current_standard.match_definition(standard_event, event)) for event in events
                    )
                    index += 1

                if not events_implemented:
                    continue    # if even one of the events was not implemented, then check the next standard

                self.standards.append(standard)
                break



from boa3.internal.model.builtin.interop.nativecontract import PolicyContractMethod
from boa3.internal.model.builtin.interop.policy.transactionattributetypetype import TransactionAttributeTypeType
from boa3.internal.model.variable import Variable


class GetAttributeFeeMethod(PolicyContractMethod):

    def __init__(self, tx_attribute_type: TransactionAttributeTypeType):
        from boa3.internal.model.type.type import Type
        identifier = 'get_attribute_fee'
        native_identifier = 'getAttributeFee'
        args: dict[str, Variable] = {
            'attribute_type': Variable(tx_attribute_type)
        }
        super().__init__(identifier, native_identifier, args, return_type=Type.int)



from boa3.internal.model.builtin.builtinproperty import IBuiltinProperty
from boa3.internal.model.builtin.interop.interopmethod import InteropMethod
from boa3.internal.model.variable import Variable


class GetInvocationCounterMethod(InteropMethod):
    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = '-get_invocation_counter'
        syscall = 'System.Runtime.GetInvocationCounter'
        args: dict[str, Variable] = {}
        super().__init__(identifier, syscall, args, return_type=Type.int)


class InvocationCounterProperty(IBuiltinProperty):
    def __init__(self):
        identifier = 'invocation_counter'
        getter = GetInvocationCounterMethod()
        super().__init__(identifier, getter)



from boa3.internal.model.builtin.interop.nativecontract import PolicyContractMethod
from boa3.internal.model.variable import Variable


class GetFeePerByteMethod(PolicyContractMethod):

    def __init__(self):
        from boa3.internal.model.type.type import Type
        identifier = 'get_fee_per_byte'
        native_identifier = 'getFeePerByte'
        args: dict[str, Variable] = {}
        super().__init__(identifier, native_identifier, args, return_type=Type.int)
